/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "Database.h"
#include "version.h"
#include "DefaultValues.h"
#include "DefinedCharacters.h"

// Qt includes
#include <QDebug>
#include <QSqlQuery>
#include <QMessageBox>
#include <QDir>
#include <QSqlError>
#include <QLockFile>
#include <QCryptographicHash>

// C++ includes
#include <algorithm> // std::sort

namespace {

// The database version we're using
constexpr int DB_VERSION = 6;

// Key literals for the database's "config" table
const QString DB_MUCKTURNIER_VERSION    = QString::fromUtf8("MUCKTURNIER_VERSION");
const QString DB_DB_VERSION             = QString::fromUtf8("DB_VERSION");
const QString DB_TOURNAMENT_MODE        = QString::fromUtf8("TOURNAMENT_MODE");
const QString DB_BOOGER_SCORE           = QString::fromUtf8("BOOGER_SCORE");
const QString DB_BOOGERS_PER_ROUND      = QString::fromUtf8("BOOGERS_PER_ROUND");
const QString DB_NO_TABLE_NUMBERS       = QString::fromUtf8("NO_TABLE_NUMBERS");
const QString DB_AUTO_SELECT_PAIRS      = QString::fromUtf8("AUTO_SELECT_PAIRS");
const QString DB_INCLUDE_OPPONENT_GOALS = QString::fromUtf8("INCLUDE_OPPONENT_GOALS");
// This one is only used to check if we can write to the database when opening it
// and is removed again afterwards, so it does never show up in a database in use.
const QString DB_WRITE_TEST             = QString::fromUtf8("WRITE_TEST");

// Values for DB_TOURNAMENT_MODE
const QString DB_FIXED_PAIRS    = QString::fromUtf8("FIXED_PAIRS");
const QString DB_SINGLE_PLAYERS = QString::fromUtf8("SINGLE_PLAYERS");

// Literal bool values
const QString DB_BOOL_FALSE = QString::fromUtf8("0");
const QString DB_BOOL_TRUE  = QString::fromUtf8("1");

// Internally used connection name
const QString DB_CONNECTION_NAME = QString::fromUtf8("SQLiteDB");

}

Database::Database(QObject *parent) : QObject(parent)
{
    m_sorter.setCaseSensitivity(Qt::CaseInsensitive);
}

Database::~Database()
{
    saveDisplayOptions();
    closeDatabaseConnection();
}

bool Database::startTransaction()
{
    bool success = m_db.transaction();
    if (! success) {
        emitSqlError();
    }
    return success;
}

bool Database::commitTransaction()
{
    bool success = m_db.commit();
    if (! success) {
        emitSqlError();
    }
    return success;
}

bool Database::queryExec(QSqlQuery &query)
{
    bool success = query.exec();
    if (! success) {
        emitSqlError(query);
    }
    return success;
}

bool Database::queryExec(QSqlQuery &query, const QString &text)
{
    bool success = query.exec(text);
    if (! success) {
        emitSqlError(query);
    }
    return success;
}

bool Database::queryExecBatch(QSqlQuery &query)
{
    bool success = query.execBatch();
    if (! success) {
        emitSqlError(query);
    }
    return success;
}

void Database::emitSqlError()
{
    QSqlError error = m_db.lastError();
    emit sqlError(tr("<p>%1<br/>%2</p>").arg(error.databaseText(), error.driverText()));
}

void Database::emitSqlError(const QSqlQuery &query)
{
    QSqlError error = query.lastError();
    emit sqlError(tr("<p>%1<br/>%2</p>").arg(error.databaseText(), error.driverText()),
                    query.executedQuery());
}

bool Database::updateValue(QSqlQuery &query, const QString &key, bool state)
{
    query.prepare(QString::fromUtf8("UPDATE config SET value = ? WHERE key = ?"));
    query.bindValue(0, state ? DB_BOOL_TRUE : DB_BOOL_FALSE);
    query.bindValue(1, key);
    return queryExec(query);
}

void Database::closeDatabaseConnection()
{
    if (! m_db.isOpen()) {
        return;
    }

    // Close the existing connection
    m_db.close();
    m_db = QSqlDatabase::database();
    QSqlDatabase::removeDatabase(DB_CONNECTION_NAME);

    // Remove the lock file
    delete m_lockFile;
}

void Database::saveDisplayOptions()
{
    if (! m_isWritable || ! m_db.isOpen()) {
        return;
    }

    startTransaction();
    QSqlQuery query(m_db);
    updateValue(query, DB_NO_TABLE_NUMBERS, m_noTableNumbers);
    updateValue(query, DB_AUTO_SELECT_PAIRS, m_autoSelectPairs);
    updateValue(query, DB_INCLUDE_OPPONENT_GOALS, m_includeOpponentGoals);
    commitTransaction();
}

bool Database::connectDatabase(const QString &dbFile)
{
    closeDatabaseConnection();

    // Create a Lockfile

    QString lockFileName = QDir::temp().absoluteFilePath(
        QString::fromUtf8("muckturnier.%1.lock").arg(
                          QString::fromUtf8(QCryptographicHash::hash(dbFile.toUtf8(),
                          QCryptographicHash::Md5).toHex()))
    );

    m_lockFile = new QLockFile(lockFileName);
    m_lockFile->setStaleLockTime(0);

    if (! m_lockFile->tryLock(0)) {
        m_errorMessage = tr("<p>Die Datenbank ist bereits von einer anderen Instanz von "
                            "Muckturnier geöffnet.</p>"
                            "<p>Sollte dies nicht der Fall sein (evtl. nach einen Crash, "
                            "Stromausfall etc.), dann bitte die Lock-Datei „%1“ manuell löschen "
                            "und die Datenbank erneut öffnen.</p>").arg(
                            QDir::toNativeSeparators(lockFileName));
        return false;
    }

    // Connect the database

    m_db = QSqlDatabase::addDatabase(QString::fromUtf8("QSQLITE"), DB_CONNECTION_NAME);
    m_db.setDatabaseName(dbFile);

    if (! m_db.open()) {
        // This should not happen
        QSqlError error = m_db.lastError();
        m_errorMessage = tr("Unerwarteter Fehler: Konnte keine Datenbankverbindung herstellen.</p>"
                            "<p>Die Fehlermeldung war: %2</p>").arg(error.text());
        return false;
    }

    return true;
}

bool Database::create(const QString &dbFile)
{
    if (! connectDatabase(dbFile)) {
        return false;
    }

    startTransaction();
    QSqlQuery query(m_db);
    queryExec(query, QString::fromUtf8("CREATE TABLE config(\n"
                                       "    key TEXT,\n"
                                       "    value TEXT\n"
                                       ")"));
    queryExec(query, QString::fromUtf8("CREATE TABLE players(\n"
                                       "    id INTEGER PRIMARY KEY,\n"
                                       "    name TEXT\n"
                                       ")"));
    queryExec(query, QString::fromUtf8("CREATE TABLE assignment(\n"
                                       "    round INTEGER,\n"
                                       "    pair_or_player INTEGER,\n"
                                       "    pair_number INTEGER,\n"
                                       "    player_number INTEGER,\n"
                                       "    table_number INTEGER\n"
                                       ")"));
    queryExec(query, QString::fromUtf8("CREATE TABLE score(\n"
                                       "    round INTEGER,\n"
                                       "    table_number INTEGER,\n"
                                       "    booger INTEGER,\n"
                                       "    pair_or_player INTEGER,\n"
                                       "    score INTEGER\n"
                                       ")"));

    query.prepare(QString::fromUtf8("INSERT INTO config(key, value) VALUES(?, ?)"));

    QVariantList keys;
    keys << DB_MUCKTURNIER_VERSION
         << DB_DB_VERSION
         << DB_TOURNAMENT_MODE
         << DB_BOOGER_SCORE
         << DB_BOOGERS_PER_ROUND
         << DB_NO_TABLE_NUMBERS
         << DB_AUTO_SELECT_PAIRS
         << DB_INCLUDE_OPPONENT_GOALS;
    query.addBindValue(keys);

    QVariantList values;
    values << QString::fromUtf8(MT_VERSION)
           << DB_VERSION
           << DB_FIXED_PAIRS
           << DEFAULT_BOOGER_SCORE
           << DEFAULT_BOOGERS_PER_ROUND
           << DB_BOOL_FALSE
           << DB_BOOL_TRUE
           << DB_BOOL_FALSE;
    query.addBindValue(values);
    queryExecBatch(query);

    commitTransaction();

    updateSettings();
    m_isWritable = true;
    m_tournamentStarted = false;

    return true;
}

void Database::updateSettings()
{
    // Reads the settings stored in the database.

    QSqlQuery query(m_db);
    queryExec(query, QString::fromUtf8("SELECT key, value FROM config"));

    while (query.next()) {
        QString key = query.value(0).toString();
        QString value = query.value(1).toString();

        if (key == DB_MUCKTURNIER_VERSION) {
            m_dbValueMuckturnierVersion = value;
        } else if (key == DB_TOURNAMENT_MODE) {
            if (value == DB_FIXED_PAIRS) {
                m_settingsTournamentMode = TournamentMode::fixedPairs;
            } else if (value == DB_SINGLE_PLAYERS) {
                m_settingsTournamentMode = TournamentMode::singlePlayers;
            }
        } else if (key == DB_BOOGER_SCORE) {
            m_settingsBoogerScore = value.toInt();
        } else if (key == DB_BOOGERS_PER_ROUND) {
            m_settingsBoogersPerRound = value.toInt();
        } else if (key == DB_NO_TABLE_NUMBERS) {
            m_noTableNumbers = value == DB_BOOL_FALSE ? false : true;
        } else if (key == DB_AUTO_SELECT_PAIRS) {
            m_autoSelectPairs = value == DB_BOOL_FALSE ? false : true;
        } else if (key == DB_INCLUDE_OPPONENT_GOALS) {
            m_includeOpponentGoals = value == DB_BOOL_FALSE ? false : true;
        }
    }
}

bool Database::tournamentStarted() const
{
    return m_tournamentStarted;
}

Database::Error Database::checkDatabase(const QString &dbFile)
{
    // Step 1: Check if it's a file and if we can read it

    QFileInfo fileInfo = QFileInfo(dbFile);
    if (! fileInfo.exists()) {
        m_errorMessage = tr("<p>Die Datei „%1“ konnte nicht gefunden werden</p>").arg(
                             QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }
    if (! (fileInfo.isFile() || fileInfo.isSymLink())) {
        m_errorMessage = tr("<p>„%1“ ist weder eine Datei, noch ein symbolischer Link auf eine "
                            "Datei</p>").arg(QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }
    if (! fileInfo.isReadable()) {
        m_errorMessage = tr("<p>Die Datei „%1“ ist nicht lesbar. Bitte die Zugriffsrechte "
                            "überprüfen!</p>").arg(QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }

    // Step 2: Check if we can connect the database and if it's actually an SQLite database

    if (! connectDatabase(dbFile)) {
        return Error::CouldNotOpen;
    }

    QSqlQuery query(m_db);

    query.exec(QString::fromUtf8("PRAGMA quick_check"));
    QString testResult;
    if (query.next()) {
        testResult = query.value(0).toString();
    }

    if (testResult != QString::fromUtf8("ok")) {
        QSqlError error = query.lastError();
        m_errorMessage = tr("<p>Die Datei „%1“ ist entweder keine SQLite-Datenbank oder eine "
                            "beschädigte SQLite-Datenbank. Die Fehlermeldung war:</p>"
                            "<p>%2</p>").arg(QDir::toNativeSeparators(dbFile), error.text());
        return Error::CouldNotOpen;
    }

    // Step 3: Check if it's a Muckturnier database

    // Let's see if we have all tables
    queryExec(query, QString::fromUtf8("SELECT name FROM sqlite_master WHERE type = 'table' "
                                           "AND name = 'config' "
                                           "OR name = 'players' "
                                           "OR name = 'assignment' "
                                           "OR name = 'score'"));
    int tableCount = 0;
    while (query.next()) {
        tableCount++;
    }
    if (tableCount != 4) {
        m_errorMessage = tr("<p>In der Datenbank „%1“ sind nicht alle nötigen Tabellen vorhanden. "
                            "Entweder ist die Datei keine Muckturnier-Datenbank, oder sie wurde "
                            "mit einer zu alten Version von Muckturnier (vor 0.5.1) erstellt."
                            "</p>").arg(QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }

    // Let's see if we have the correct database version

    query.prepare(QString::fromUtf8("SELECT value FROM config WHERE key = ?"));
    query.bindValue(0, DB_DB_VERSION);
    queryExec(query);

    int dbVersion = 0;
    if (query.next()) {
        dbVersion = query.value(0).toInt();
    }

    if (dbVersion < 3) {
        m_errorMessage = tr("<p>Es wurde in der Datenbank „%1“ kein gültiger Konfigurationseintrag "
                            "zur verwendeten Datenbankrevision gefunden. Die Datei ist entweder "
                            "keine Muckturnier-Datenbank oder sie wurde mit einer zu alten Version "
                            "von Muckturnier (vor 0.5.1) erstellt.</p>").arg(
                            QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    } else if (dbVersion > DB_VERSION) {
        m_errorMessage = tr("<p>Die Datenbank „%1“ hat die falsche Revision (ist %2, die benutzte "
                            "Version von Muckturnier nutzt %3). Sie wurde vermutlich mit einer "
                            "neueren Version erstellt.</p>"
                            "<p>Bitte auf die neueste Version aktualisieren (zu finden unter "
                            "<a href=\"http://muckturnier.org/\">http://muckturnier.org/</a>) und "
                            "die Datenbank damit erneut öffnen.</p>").arg(
                            QDir::toNativeSeparators(dbFile), QString::number(dbVersion),
                                                              QString::number(DB_VERSION));
        return Error::CouldNotOpen;
    } else if (dbVersion == 3) {
        return Error::DbVersion3;
    } else if (dbVersion == 4) {
        return Error::DbVersion4;
    } else if (dbVersion == 5) {
        return Error::DbVersion5;
    }

    // If we reach here, we actually can assume that this is a valid Muckturnier database ;-)
    closeDatabaseConnection();
    return Error::NoError;
}

Database::Error Database::openDb(const QString &dbFile)
{
    Error testResult = checkDatabase(dbFile);
    if (testResult != Error::NoError) {
        return testResult;
    }

    connectDatabase(dbFile);
    updateSettings();
    checkIfTournamentStarted();

    // We actually have to check if a database is writable by actually trying to insert some data
    // (is this a bug or a feature?!) because:
    //
    // - If the file itself has write access, QFileInfo::isWritable() will return true, but if the
    //   parent directory is not writable, we won't be able to write to the datbase.
    //
    // - Checking for the parent directory to be writable in addition does also not work, as we will
    //   get false for "My Documents" on Windows.
    //
    // - Opening the file via QFile::open() does actually work if the file is writable and the
    //   parent directory is not (with the QIODevice::WriteOnly flag, the file will even be
    //   trucated!), but we still can't insert data in this case.
    //
    // So here we go:

    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("INSERT INTO config(key, value) VALUES(?, ?)"));
    query.bindValue(0, DB_WRITE_TEST);
    query.bindValue(1, QString::fromUtf8("test"));

    if (query.exec()) {
        query.prepare(QString::fromUtf8("DELETE FROM config WHERE key = ?"));
        query.bindValue(0, DB_WRITE_TEST);
        query.exec();
        m_db.commit();
        m_isWritable = true;
        return Error::NoError;
    } else {
        m_db.commit();
        m_isWritable = false;
        return Error::NoErrorReadOnly;
    }
}

QString Database::errorMessage() const
{
    return m_errorMessage;
}

QString Database::playersString(Numerus numerus, bool article) const
{
    switch (m_settingsTournamentMode) {
    case TournamentMode::fixedPairs:
        switch (numerus) {
        case Numerus::singular:
            return article ? tr("das Paar")  : tr("Paar");
        case Numerus::plural:
            return article ? tr("die Paare") : tr("Paare");
        }
    case TournamentMode::singlePlayers:
        switch (numerus) {
        case Numerus::singular:
            return article ? tr("der Spieler") : tr("Spieler");
        case Numerus::plural:
            return article ? tr("die Spieler") : tr("Spieler");
        }
    }
    // This is just to make the compiler happy. We can't reach here.
    return QString();
}

void Database::reset()
{
    startTransaction();
    QSqlQuery query(m_db);
    queryExec(query, QString::fromUtf8("DELETE FROM players"));
    queryExec(query, QString::fromUtf8("DELETE FROM assignment"));
    queryExec(query, QString::fromUtf8("DELETE FROM score"));
    commitTransaction();
    m_tournamentStarted = false;
}

void Database::saveSettings()
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("UPDATE config SET value = ? WHERE key = ?"));

    QVariantList values;
    switch (m_settingsTournamentMode) {
    case TournamentMode::fixedPairs:
        values << DB_FIXED_PAIRS;
        break;
    case TournamentMode::singlePlayers:
        values << DB_SINGLE_PLAYERS;
        break;
    }
    values << m_settingsBoogerScore
           << m_settingsBoogersPerRound;
    query.addBindValue(values);

    QVariantList keys;
    keys << DB_TOURNAMENT_MODE
         << DB_BOOGER_SCORE
         << DB_BOOGERS_PER_ROUND;
    query.addBindValue(keys);

    queryExecBatch(query);
    commitTransaction();
}

void Database::checkIfTournamentStarted()
{
    QSqlQuery query(m_db);
    queryExec(query, QString::fromUtf8("SELECT COUNT(round) FROM SCORE"));
    query.next();
    m_tournamentStarted = query.value(0).toInt() != 0;
}

int Database::necessaryDivisor() const
{
    return 2 + (m_settingsTournamentMode != TournamentMode::fixedPairs) * 2;
}

QStringList Database::listPlayers()
{
    QStringList playersList;
    QSqlQuery query(m_db);
    queryExec(query, QString::fromUtf8("SELECT name FROM players"));
    while (query.next()) {
        playersList << query.value(0).toString();
    }
    return playersList;
}

bool Database::checkName(QString name)
{
    // Checks if a player/pair name already exists.
    name = name.toLower().remove(MARKER_CHARACTER);
    for (const QString &player : listPlayers()) {
        if (player.toLower().remove(MARKER_CHARACTER) == name) {
            return false;
        }
    }
    return true;
}

void Database::registerPlayers(const QString &name)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("INSERT INTO players(id, name) VALUES(NULL, ?)"));
    query.bindValue(0, name);
    queryExec(query);
    commitTransaction();
}

void Database::registerPlayers(const QVariantList &names)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("INSERT INTO players(id, name) VALUES(NULL, ?)"));
    query.addBindValue(names);
    queryExecBatch(query);
    commitTransaction();
}

void Database::deletePlayers(const QString &name)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("DELETE FROM players WHERE name = ?"));
    query.bindValue(0, name);
    queryExec(query);
    commitTransaction();
}

void Database::deletePlayers(const QVariantList &names)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("DELETE FROM players WHERE name = ?"));
    query.addBindValue(names);
    queryExecBatch(query);
    commitTransaction();
}

void Database::renamePlayers(const QString &oldName, const QString &newName)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("UPDATE players SET name = ? WHERE name = ?"));
    query.bindValue(0, newName);
    query.bindValue(1, oldName);
    queryExec(query);
    commitTransaction();
}

void Database::renamePlayers(const QVariantList &oldNames, const QVariantList &newNames)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("UPDATE players SET name = ? WHERE name = ?"));
    query.addBindValue(newNames);
    query.addBindValue(oldNames);
    queryExecBatch(query);
    commitTransaction();
}

int Database::playersCount()
{
    QSqlQuery query(m_db);
    queryExec(query, QString::fromUtf8("SELECT COUNT(id) AS count FROM players"));
    query.next();
    return query.value(0).toInt();
}


QMap<int, bool> Database::roundsState()
{
    QMap<int, bool> roundsStateMap;
    int allBoogers = playersCount() * m_settingsBoogersPerRound;
    int roundNumber = 1;

    QSqlQuery query(m_db);
    queryExec(query,QString::fromUtf8("SELECT round, COUNT(round) AS count FROM score "
                                      "GROUP BY round ORDER BY round"));

    // Check if each round already contains all boogers
    while (query.next()) {
        roundNumber = query.value(0).toInt();
        if (query.value(1).toInt() < allBoogers) {
            roundsStateMap[roundNumber] = false;
        } else {
            roundsStateMap[roundNumber] = true;
        }
    }

    // Be sure to have a state for each round. All data from a round could have been deleted.
    // If we have no data, round 1 will be added as open.
    for (int i = 1; i <= roundNumber; i++) {
        if (! roundsStateMap.contains(i)) {
            roundsStateMap[i] = false;
        }
    }

    return roundsStateMap;
}

QList<int> Database::missingTables(int round)
{
    QList<int> missingTableNumbers;

    // First, we assume we need all tables
    for (int i = 1; i <= playersCount() / necessaryDivisor(); i++) {
        missingTableNumbers << i;
    }

    // Then, we take away those that already have a score
    QSqlQuery query(m_db);
    const QString queryString = QString::fromUtf8("SELECT table_number FROM score "
                                                  "WHERE round = ? "
                                                  "GROUP BY table_number");
    query.prepare(queryString);
    query.bindValue(0, round);
    queryExec(query);
    while (query.next()) {
        missingTableNumbers.removeOne(query.value(0).toInt());
    }

    return missingTableNumbers;
}

QMap<QString, int> Database::missingPlayers(int round)
{
    QMap<QString, int> missingPlayersMap;
    QSqlQuery query(m_db);

    const QString queryString = QString::fromUtf8(
        "SELECT players.name, players.id "
        "FROM players "
        "WHERE NOT EXISTS ( "
            "SELECT score.pair_or_player "
            "FROM score "
            "WHERE players.id = score.pair_or_player "
            "AND round = ? "
        ")"
    );
    query.prepare(queryString);
    query.bindValue(0, round);
    queryExec(query);
    while (query.next()) {
        missingPlayersMap[query.value(0).toString()] = query.value(1).toInt();
    }

    return missingPlayersMap;
}

Database::Assignment Database::getAssignment(int roundNumber)
{
    int round = roundNumber - 1;
    int allTables = playersCount() / necessaryDivisor();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("SELECT pair_or_player, pair_number, table_number "
                                    "FROM assignment WHERE round = ?"));
    query.bindValue(0, round);
    queryExec(query);

    QMap<int, int> assignedTable;
    QMap<int, int> assignedPairNumber;
    QMap<int, QMap<int, int>> assignedPair;

    while(query.next()) {
        if (query.value(1).toInt() == 1) {
            assignedTable[query.value(0).toInt()] = query.value(2).toInt();
        } else {
            assignedTable[query.value(0).toInt()] = query.value(2).toInt() % allTables + 1;
        }
        assignedPairNumber[query.value(0).toInt()] = query.value(1).toInt();
    }

    for (const int &pairNumber : assignedTable.keys()) {
        assignedPair[assignedTable[pairNumber]][assignedPairNumber[pairNumber]] = pairNumber;
    }

    Assignment assignment;
    assignment.assignedTable = assignedTable;
    assignment.assignedPairNumber = assignedPairNumber;
    assignment.assignedPair = assignedPair;

    return assignment;
}

QList<Database::Score> Database::getScores(int round)
{
    QSqlQuery query(m_db);
    switch (m_settingsTournamentMode) {
    case TournamentMode::fixedPairs:
        query.prepare(QString::fromUtf8(
            "SELECT "
                "score.table_number, "
                "players.name, "
                "score.score "
            "FROM score "
            "INNER JOIN players ON "
                "players.id = score.pair_or_player "
            "INNER JOIN assignment ON "
                "assignment.round = score.round "
                "AND assignment.pair_or_player = score.pair_or_player "
            "WHERE score.round = ? "
            "ORDER BY "
                "score.table_number, "
                "assignment.pair_number, "
                "score.booger"
        ));
        break;
    case TournamentMode::singlePlayers:
        // It's a bit more complicated for single players ;-)
        query.prepare(QString::fromUtf8(
            "SELECT "
                "score_one.table_number, "
                "player_one.name || ' / ' || player_two.name, "
                "score_one.score "
            "FROM "
                "score AS score_one "
            "INNER JOIN players AS player_one ON "
                "player_one.id = score_one.pair_or_player "
            "INNER JOIN assignment AS assignment_one ON "
                "assignment_one.round = score_one.round "
                "AND assignment_one.table_number = score_one.table_number "
                "AND assignment_one.pair_or_player = score_one.pair_or_player "
            "INNER JOIN players AS player_two ON "
                "player_two.id = score_two.pair_or_player "
            "INNER JOIN score AS score_two ON "
                "score_two.round = score_one.round "
                "AND score_two.table_number = score_one.table_number "
                "AND score_two.booger = score_one.booger "
                "AND assignment_two.pair_number = assignment_one.pair_number "
                "AND assignment_two.player_number = 2 "
            "INNER JOIN assignment AS assignment_two ON "
                "assignment_two.round = score_one.round "
                "AND assignment_two.table_number = score_one.table_number "
                "AND assignment_two.pair_or_player = score_two.pair_or_player "
            "WHERE "
                "score_one.round = ? "
                "AND assignment_one.player_number = 1 "
            "ORDER BY "
                "score_one.table_number, "
                "assignment_one.pair_number, "
                "score_one.booger "
        ));
        break;
    }
    query.bindValue(0, round);
    queryExec(query);

    QList<Score> scores;
    while (query.next()) {
        Score score;
        score.table = query.value(0).toInt();
        score.pairName = query.value(1).toString();
        score.score = query.value(2).toInt();
        scores << score;
    }

    return scores;
}

QList<Database::SingleScore> Database::getSingleScore(int round, int table)
{
    QList<SingleScore> score;
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8(
        "SELECT "
            "score.pair_or_player, "
            "score.score "
        "FROM score "
        "INNER JOIN assignment ON "
            "assignment.round = score.round "
            "AND assignment.pair_or_player = score.pair_or_player "
        "WHERE "
            "score.round = ? "
            "AND score.table_number = ? "
        "ORDER BY "
            "assignment.pair_number, "
            "assignment.player_number, "
            "score.booger"
    ));
    query.bindValue(0, round);
    query.bindValue(1, table);
    queryExec(query);

    while (query.next()) {
        SingleScore singleScore;
        singleScore.pairOrPlayer = query.value(0).toInt();
        singleScore.score = query.value(1).toInt();
        score << singleScore;
    }

    return score;
}

void Database::insertScore(QSqlQuery &query,
                           int round, int table, int booger, int pairOrPlayer, int score)
{
    query.prepare(QString::fromUtf8("INSERT INTO score(round, table_number, booger, "
                                                      "pair_or_player, score) "
                                    "VALUES(?, ?, ?, ?, ?)"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.bindValue(2, booger);
    query.bindValue(3, pairOrPlayer);
    query.bindValue(4, score);
    queryExec(query);
}

void Database::updateAssignment(QSqlQuery &query, int round, int pairOrPlayer,
                                                  int pairNumber, int playerNumber, int table)
{
    query.prepare(QString::fromUtf8(
        "INSERT INTO assignment(round, pair_or_player, pair_number, player_number, table_number) "
        "VALUES(?, ?, ?, ?, ?)"
    ));
    query.bindValue(0, round);
    query.bindValue(1, pairOrPlayer);
    query.bindValue(2, pairNumber);
    query.bindValue(3, playerNumber);
    query.bindValue(4, table);
    queryExec(query);
}

void Database::saveScore(int round, int table, const QList<int> &players,
                         const QList<int> &pair1Scores, const QList<int> &pair2Scores)
{
    startTransaction();
    QSqlQuery query(m_db);

    for (int booger = 1; booger <= pair1Scores.size(); booger++) {
        insertScore(query, round, table, booger, players.at(0), pair1Scores.at(booger - 1));
        insertScore(query, round, table, booger, players.at(1), pair2Scores.at(booger - 1));
        if (m_settingsTournamentMode == TournamentMode::singlePlayers) {
            insertScore(query, round, table, booger, players.at(2), pair1Scores.at(booger - 1));
            insertScore(query, round, table, booger, players.at(3), pair2Scores.at(booger - 1));
        }

        if (booger == 1) {
            updateAssignment(query, round, players.at(0), 1, 1, table);
            updateAssignment(query, round, players.at(1), 2, 1, table);
            if (m_settingsTournamentMode == TournamentMode::singlePlayers) {
                updateAssignment(query, round, players.at(2), 1, 2, table);
                updateAssignment(query, round, players.at(3), 2, 2, table);
            }
        }
    }

    commitTransaction();

    if (! m_tournamentStarted) {
        m_tournamentStarted = true;
    }
}

void Database::deleteScore(int round, int table)
{
    startTransaction();
    QSqlQuery query(m_db);

    query.prepare(QString::fromUtf8("DELETE FROM score WHERE round = ? AND table_number = ?"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    queryExec(query);

    query.prepare(QString::fromUtf8("DELETE FROM assignment WHERE round = ? AND table_number = ?"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    queryExec(query);

    commitTransaction();
    checkIfTournamentStarted();
}

QList<Database::Ranking> Database::getRanking(int upToRound)
{
    QSqlQuery query(m_db);
    QString queryString = QString::fromUtf8(
        "SELECT "
            "players.name, "
            "SUM(CASE WHEN one_score.score = ? THEN 1 ELSE 0 END) AS won_boogers, "
            "SUM(CASE WHEN one_score.score = ? THEN 0 ELSE one_score.score END) AS scored_goals, "
            "SUM(CASE WHEN other_score.score = ? THEN 0 ELSE other_score.score END) AS opponent_goals, "
            "COUNT(one_score.score) AS played_boogers "
        "FROM "
            "score AS one_score "
        "INNER JOIN players ON "
            "players.id = one_score.pair_or_player "
        "INNER JOIN assignment AS one_assignment ON "
            "one_assignment.round = one_score.round "
            "AND one_assignment.table_number = one_score.table_number "
            "AND one_assignment.pair_or_player = one_score.pair_or_player "
        "INNER JOIN assignment AS other_assignment ON "
            "other_assignment.round = other_score.round "
            "AND other_assignment.table_number = other_score.table_number "
            "AND other_assignment.pair_or_player = other_score.pair_or_player "
        "INNER JOIN score AS other_score ON "
            "other_score.round = one_score.round "
            "AND other_score.table_number = one_score.table_number "
            "AND other_score.booger = one_score.booger "
            "AND other_assignment.pair_number != one_assignment.pair_number "
        "WHERE one_score.round <= ? "
        "GROUP BY "
            "one_score.pair_or_player "
        "ORDER BY "
            "won_boogers DESC, "
            "scored_goals DESC, "
            "opponent_goals, "
            "players.name"
    );
    query.prepare(queryString);
    for (int i = 0; i < 3; i++) {
        query.bindValue(i, m_settingsBoogerScore);
    }
    query.bindValue(3, upToRound);
    queryExec(query);

    QList<Ranking> allRankings;
    while (query.next()) {
        Ranking ranking;
        ranking.name = query.value(0).toString();
        ranking.boogers = query.value(1).toInt();
        ranking.goals = query.value(2).toInt();
        ranking.opponentGoals = query.value(3).toInt();
        ranking.playedBoogers = query.value(4).toInt();
        if (m_settingsTournamentMode == TournamentMode::singlePlayers) {
            ranking.boogers /= 2;
            ranking.goals /= 2;
            ranking.opponentGoals /= 2;
            ranking.playedBoogers /= 2;
        }
        allRankings << ranking;
    }

    if (allRankings.count() < playersCount()) {
        // Some pair(s)/player(s) don't have a score yet,
        // so we insert empty scores to get a complete ranking list

        queryExec(query, QString::fromUtf8("SELECT players.name "
                                           "FROM players "
                                           "WHERE NOT EXISTS ( "
                                               "SELECT score.pair_or_player "
                                               "FROM score "
                                               "WHERE players.id = score.pair_or_player "
                                           ")"));

        QStringList missingNames;
        while (query.next()) {
            missingNames << query.value(0).toString();
        }
        sortList(missingNames);

        Ranking missingRanking;
        missingRanking.boogers = 0;
        missingRanking.goals = 0;
        missingRanking.opponentGoals = 0;
        missingRanking.playedBoogers = 0;
        for (const QString &name : missingNames) {
            missingRanking.name = name;
            allRankings << missingRanking;
        }
    }

    return allRankings;
}

QString Database::getMuckturnierVersion() const
{
    return m_dbValueMuckturnierVersion;
}

Database::TournamentMode Database::getTournamentMode() const
{
    return m_settingsTournamentMode;
}

QString Database::getDbFile() const
{
    return m_db.databaseName();
}

void Database::setTournamentMode(Database::TournamentMode tournamentMode)
{
    m_settingsTournamentMode = tournamentMode;
}

int Database::getDbVersion() const
{
    return DB_VERSION;
}

int Database::getBoogerScore() const
{
    return m_settingsBoogerScore;
}

void Database::setBoogerScore(int boogerScore)
{
    m_settingsBoogerScore = boogerScore;
}

int Database::getBoogersPerRound() const
{
    return m_settingsBoogersPerRound;
}

void Database::setBoogersPerRound(int boogersPerRound)
{
    m_settingsBoogersPerRound = boogersPerRound;
}

QString Database::boogify(int score)
{
    if (score == m_settingsBoogerScore) {
        return QString::number(score);
    } else {
        return QString::fromUtf8("%1 %2").arg(QString::number(score), BOOGER_CHARACTER);
    }
}

void Database::updateDatabase3to4(const QString &dbFileName)
{
    connectDatabase(dbFileName);

    startTransaction();
    QSqlQuery query(m_db);

    // Get the tournament mode
    queryExec(query, QString::fromUtf8("SELECT value FROM config WHERE key = 'TOURNAMENT_MODE'"));
    query.next();
    QString tournamentMode = query.value(0).toString();

    // Add the missing column to "assignment"
    queryExec(query, QString::fromUtf8("ALTER TABLE assignment ADD COLUMN player_number INT"));

    // Fill in the data. For a single player mode database, the fetch result hopefully is still in
    // the correct order, so that the right players become player 1 and 2 inside a team.
    if (tournamentMode == QString::fromUtf8("FIXED_PAIRS")) {
        queryExec(query, QString::fromUtf8("UPDATE assignment SET player_number = 1"));
    } else {
        QVariantList round;
        QVariantList pair_or_player;
        QVariantList pair_number;
        QVariantList table_number;
        QVariantList player_number;
        queryExec(query, QString::fromUtf8("SELECT round, pair_or_player, pair_number, "
                                                  "table_number "
                                           "FROM assignment "
                                           "ORDER BY round, table_number, pair_number"));
        int playerNumber = 1;
        while (query.next()) {
            round << query.value(0).toInt();
            pair_or_player << query.value(1).toInt();
            pair_number << query.value(2).toInt();
            table_number << query.value(3).toInt();
            playerNumber = (playerNumber + 1) % 2;
            player_number << playerNumber + 1;
        }

        query.prepare(QString::fromUtf8("UPDATE assignment SET player_number = ? WHERE "
                                        "round = ? "
                                        "AND pair_or_player = ? "
                                        "AND pair_number = ? "
                                        "AND table_number = ?"));
        query.addBindValue(player_number);
        query.addBindValue(round);
        query.addBindValue(pair_or_player);
        query.addBindValue(pair_number);
        query.addBindValue(table_number);
        queryExecBatch(query);
    }

    // Finally, we can update the "DB_VERSION" key.
    queryExec(query, QString::fromUtf8("UPDATE config SET value = 4 WHERE key = 'DB_VERSION'"));

    commitTransaction();
    closeDatabaseConnection();
}

void Database::updateDatabase4to5(const QString &dbFileName)
{
    connectDatabase(dbFileName);

    startTransaction();
    QSqlQuery query(m_db);

    // Get the boogers per round
    queryExec(query, QString::fromUtf8("SELECT value FROM config WHERE key = 'GAMES_PER_ROUND'"));
    query.next();
    QString boogersPerRound = query.value(0).toString();

    // Update "GAMES_PER_ROUND" to "BOOGERS_PER_ROUND"
    queryExec(query, QString::fromUtf8("DELETE FROM config WHERE key = 'GAMES_PER_ROUND'"));
    query.prepare(QString::fromUtf8("INSERT INTO config(key, value) VALUES('BOOGERS_PER_ROUND', "
                                                                          "?)"));
    query.addBindValue(boogersPerRound);
    queryExec(query);

    // Update the score table's "game" column to "booger"
    queryExec(query, QString::fromUtf8("ALTER TABLE score RENAME TO score_tmp"));
    queryExec(query, QString::fromUtf8("CREATE TABLE score(\n"
                                       "    round INTEGER,\n"
                                       "    table_number INTEGER,\n"
                                       "    booger INTEGER,\n"
                                       "    pair_or_player INTEGER,\n"
                                       "    score INTEGER\n"
                                       ")"));
    queryExec(query, QString::fromUtf8("INSERT INTO score(round, table_number, booger, "
                                                         "pair_or_player, score) "
                                       "SELECT round, table_number, game, pair_or_player, score "
                                       "FROM score_tmp"));
    queryExec(query, QString::fromUtf8("DROP TABLE score_tmp"));

    // Update the "DB_VERSION" key.
    queryExec(query, QString::fromUtf8("UPDATE config SET value = 5 WHERE key = 'DB_VERSION'"));

    commitTransaction();
    closeDatabaseConnection();
}

void Database::updateDatabase5to6(const QString &dbFileName)
{
    connectDatabase(dbFileName);

    startTransaction();
    QSqlQuery query(m_db);

    // Insert the new config keys. We don't use the defined constants here in case they change with
    // some later version.
    QVariantList keys;
    keys << QString::fromUtf8("NO_TABLE_NUMBERS")
         << QString::fromUtf8("AUTO_SELECT_PAIRS")
         << QString::fromUtf8("INCLUDE_OPPONENT_GOALS");
    QVariantList values;
    values << QString::fromUtf8("0")
           << QString::fromUtf8("1")
           << QString::fromUtf8("0");
    query.prepare(QString::fromUtf8("INSERT INTO config(key, value) VALUES(?, ?)"));
    query.addBindValue(keys);
    query.addBindValue(values);
    queryExecBatch(query);

    // Update the "DB_VERSION" key.
    queryExec(query, QString::fromUtf8("UPDATE config SET value = 6 WHERE key = 'DB_VERSION'"));

    commitTransaction();
    closeDatabaseConnection();
}

bool Database::isWritable() const
{
    return m_isWritable;
}

void Database::sortList(QStringList &list) const
{
    // According to the Qt Wiki, one should use the following lambda function:
    // [&](const QString &a, const QString &b) { return m_sorter.compare(a, b) < 0; }
    // but referencing the QCollator directly works just fine!
    std::sort(list.begin(), list.end(), m_sorter);
}

bool Database::noTableNumbers() const
{
    return m_noTableNumbers;
}

bool Database::autoSelectPairs() const
{
    return m_autoSelectPairs;
}

bool Database::includeOpponentGoals() const
{
    return m_includeOpponentGoals;
}

void Database::setNoTableNumbers(bool state)
{
    m_noTableNumbers = state;
}

void Database::setAutoSelectPairs(bool state)
{
    m_autoSelectPairs = state;
}

void Database::setIncludeOpponentGoals(bool state)
{
    m_includeOpponentGoals = state;
}
