/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATABASE_H
#define DATABASE_H

// Qt includes
#include <QObject>
#include <QSqlDatabase>
#include <QMap>
#include <QCollator>

// Qt classes
class QSqlQuery;
class QLockFile;

class Database : public QObject
{
    Q_OBJECT

public:
    enum Error {
        NoError,
        NoErrorReadOnly,
        CouldNotOpen,
        DbVersion3,
        DbVersion4,
        DbVersion5
    };

    enum TournamentMode {
        fixedPairs,
        singlePlayers
    };

    enum Numerus {
        singular,
        plural
    };

    struct Assignment {
        QMap<int, int> assignedTable;
        QMap<int, int> assignedPairNumber;
        QMap<int, QMap<int, int>> assignedPair;
    };

    struct Score {
        int table;
        QString pairName;
        int score;
    };

    struct SingleScore {
        int pairOrPlayer;
        int score;
    };

    struct Ranking {
        QString name;
        int boogers;
        int goals;
        int opponentGoals;
        int playedBoogers;
    };

    explicit Database(QObject *parent);
    ~Database();
    bool create(const QString &dbFile);
    Database::Error checkDatabase(const QString &dbFile);
    bool tournamentStarted() const;
    Database::Error openDb(const QString &dbFile);
    QString errorMessage() const;
    QString playersString(Numerus numerus, bool article = false) const;
    void reset();
    void saveSettings();
    void deletePlayers(const QString &name);
    void deletePlayers(const QVariantList &names);
    QStringList listPlayers();
    int necessaryDivisor() const;
    bool checkName(QString name);
    void registerPlayers(const QString &name);
    void registerPlayers(const QVariantList &names);
    void renamePlayers(const QString &oldName, const QString &newName);
    void renamePlayers(const QVariantList &oldNames, const QVariantList &newNames);
    QMap<int, bool> roundsState();
    QList<int> missingTables(int round);
    QMap<QString, int> missingPlayers(int round);
    Assignment getAssignment(int roundNumber);
    QList<Score> getScores(int round);
    QList<SingleScore> getSingleScore(int round, int table);
    QString boogify(int score);
    void saveScore(int round, int table, const QList<int> &players,
                   const QList<int> &pair1Scores, const QList<int> &pair2Scores);
    void deleteScore(int round, int table);
    QList<Ranking> getRanking(int upToRound);
    QString getMuckturnierVersion() const;
    int getDbVersion() const;
    QString getDbFile() const;
    Database::TournamentMode getTournamentMode() const;
    int getBoogerScore() const;
    int getBoogersPerRound() const;
    int playersCount();
    void setTournamentMode(Database::TournamentMode tournamentMode);
    void setBoogerScore(int boogerScore);
    void setBoogersPerRound(int boogersPerRound);
    void closeDatabaseConnection();
    void updateDatabase3to4(const QString &dbFileName);
    void updateDatabase4to5(const QString &dbFileName);
    void updateDatabase5to6(const QString &dbFileName);
    bool isWritable() const;
    void sortList(QStringList &list) const;
    bool noTableNumbers() const;
    bool autoSelectPairs() const;
    bool includeOpponentGoals() const;
    void setNoTableNumbers(bool state);
    void setAutoSelectPairs(bool state);
    void setIncludeOpponentGoals(bool state);

signals:
    void sqlError(const QString &error, const QString &query = QString());

private: // Functions
    bool connectDatabase(const QString &dbFile);
    bool startTransaction();
    bool commitTransaction();
    bool queryExec(QSqlQuery &query);
    bool queryExec(QSqlQuery &query, const QString &text);
    bool queryExecBatch(QSqlQuery &query);
    void emitSqlError();
    void emitSqlError(const QSqlQuery &query);
    void updateSettings();
    void checkIfTournamentStarted();
    void insertScore(QSqlQuery &query,
                     int round, int table, int booger, int pairOrPlayer, int score);
    void updateAssignment(QSqlQuery &query,
                          int round, int pairOrPlayer, int pairNumber, int playerNumber, int table);
    bool updateValue(QSqlQuery &query, const QString &key, bool state);
    bool lockDatabase(const QString &dbFileName);
    void saveDisplayOptions();

private: // Variables
    QSqlDatabase m_db;
    bool m_tournamentStarted;
    QString m_errorMessage;
    QString m_dbValueMuckturnierVersion;
    TournamentMode m_settingsTournamentMode;
    int m_settingsBoogerScore;
    int m_settingsBoogersPerRound;
    bool m_isWritable;
    QCollator m_sorter;
    bool m_noTableNumbers;
    bool m_autoSelectPairs;
    bool m_includeOpponentGoals;
    QLockFile *m_lockFile;

};

#endif // DATABASE_H
