/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "Dialog.h"
#include "Database.h"

// Qt includes
#include <QVBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QStandardPaths>
#include <QFileInfo>
#include <QFileDialog>
#include <QDebug>
#include <QMessageBox>
#include <QDesktopServices>

Dialog::Dialog(QWidget *parent, Database *database, DeleteMode deleteMode)
    : QDialog(parent), m_db(database)
{
    setAttribute(Qt::WA_DeleteOnClose, deleteMode == DeleteMode::DeleteOnClose);

    m_mainLayout = new QVBoxLayout;
    setLayout(m_mainLayout);

    m_title = new QLabel;
    m_title->setAlignment(Qt::AlignHCenter);
    m_title->setStyleSheet(QString::fromUtf8("QLabel { font-weight: bold; font-size: %1pt }").arg(
        (int) double(m_title->font().pointSize()) * 1.2));
    m_title->hide();
    m_mainLayout->addWidget(m_title);

    m_layout = new QVBoxLayout;
    m_mainLayout->addLayout(m_layout);

    m_buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
    m_mainLayout->addWidget(m_buttonBox);
    connect(m_buttonBox, &QDialogButtonBox::accepted, this, &Dialog::accept);
    connect(m_buttonBox, &QDialogButtonBox::rejected, this, &QDialog::reject);
}

Database * Dialog::db() const
{
    return m_db;
}

void Dialog::setTitle(const QString &title)
{
    setWindowTitle(QString::fromUtf8("Muckturnier: %1").arg(title));
    m_title->setText(title);
    m_title->show();
}

void Dialog::setOkButtonText(const QString &text)
{
    m_buttonBox->button(QDialogButtonBox::Ok)->setText(text);
}

QVBoxLayout * Dialog::layout() const
{
    return m_layout;
}

QVBoxLayout * Dialog::mainLayout() const
{
    return m_mainLayout;
}

void Dialog::setTargetFileBaseName(const QString &targetFileBaseName)
{
    m_targetFileBaseName = targetFileBaseName;
}

QString Dialog::targetFileBaseName() const
{
    return m_targetFileBaseName;
}

void Dialog::setTargetFileExtension(const QString &targetFileExtension)
{
    m_targetFileExtension = targetFileExtension;
}

void Dialog::setTargetFileDescription(const QString &targetFileDescription)
{
    m_targetFileDescription = targetFileDescription;
}

void Dialog::setFileDialogTitle(const QString &fileDialogTitle)
{
    m_fileDialogTitle = fileDialogTitle;
}

QString Dialog::getTargetFileName()
{
    QString defaultFileName = QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)
                              + QString::fromUtf8("/%1.%2").arg(m_targetFileBaseName,
                                                                m_targetFileExtension);
    int number = 0;
    while (QFileInfo::exists(defaultFileName)) {
        number++;
        defaultFileName = QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)
                          + QString::fromUtf8("/%1_%2.%3").arg(m_targetFileBaseName,
                                                               QString::number(number),
                                                               m_targetFileExtension);
    }

    QString fileName = QFileDialog::getSaveFileName(this,
        m_fileDialogTitle,
        QDir::toNativeSeparators(defaultFileName),
        QString::fromUtf8("%1 (*.%2);;").arg(m_targetFileDescription, m_targetFileExtension)
        + tr("Alle Dateien") + QString::fromUtf8(" (*)"));

    if (fileName.isEmpty()) {
        return QString();
    }

    // The user could have selected the currently open database

    if (fileName == m_db->getDbFile()) {
        QMessageBox messageBox(this);
        messageBox.setIcon(QMessageBox::Warning);
        messageBox.setWindowTitle(windowTitle());
        messageBox.setText(tr("Die ausgewählte Datei ist die momentan geöffnete Turnierdatenbank. "
                              "Soll sie wirklich überschrieben werden? Alle Daten gehen dabei "
                              "verloren!"));
        messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::Cancel);
        messageBox.setDefaultButton(QMessageBox::Cancel);

        if (messageBox.exec() == QMessageBox::Cancel) {
            return QString();
        }

        m_db->closeDatabaseConnection();
    }

    // Check if we can write the file. If it exists, it will be deleted, but the user has been
    // asked by the QFileDialog if this is okay. And yes, we have to do it in this way, because
    // QFileInfo::isWritable does not work as expected on Windows (returns false for
    // "My Documents").
    QFile test(fileName);
    if (test.open(QIODevice::WriteOnly)) {
        test.close();
        test.remove();
    } else {
        QMessageBox::critical(this,
            windowTitle(),
            tr("Es bestehen keine Schreibrechte für die Datei „%1“. "
               "Bitte die Zugriffsrechte überprüfen!").arg(QDir::toNativeSeparators(fileName)),
            QMessageBox::Ok);
        return QString();
    }

    return fileName;
}

void Dialog::successMessage(const QString &title, const QString &message)
{
    QMessageBox::information(this, title, message, QMessageBox::Ok);
}

void Dialog::openFile(const QString &fileName)
{
    QDesktopServices::openUrl(QUrl::fromLocalFile(fileName));
}

void Dialog::setOkButtonEnabled(bool state)
{
    m_buttonBox->button(QDialogButtonBox::Ok)->setEnabled(state);
}

void Dialog::showOnlyOkButton()
{
    m_buttonBox->button(QDialogButtonBox::Cancel)->hide();
}

void Dialog::accept()
{
    QDialog::accept();
}

void Dialog::setButtonText(QDialogButtonBox::StandardButton button, const QString &text)
{
    m_buttonBox->button(button)->setText(text);
}

bool Dialog::okButtonEnabled() const
{
    return m_buttonBox->button(QDialogButtonBox::Ok)->isEnabled();
}
