/* Copyright (c) 2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "ExportPlayersDialog.h"
#include "Database.h"
#include "DefinedCharacters.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QCheckBox>
#include <QFile>
#include <QTextCodec>
#include <QApplication>
#include <QDir>
#include <QRadioButton>

ExportPlayersDialog::ExportPlayersDialog(QWidget *parent, Database *database)
    : Dialog(parent, database)
{
    int playersCount = 0;
    int markedPlayersCount = 0;
    m_playersList = db()->listPlayers();
    for (const QString &name : m_playersList) {
        playersCount++;
        if (name.contains(MARKER_CHARACTER)) {
            markedPlayersCount++;
        }
    }
    bool markersSet = markedPlayersCount > 0;
    bool unMarkedPlayersThere = markedPlayersCount < playersCount;

    // Dialog settings

    setTitle(tr("%1liste exportieren").arg(db()->playersString(Database::Numerus::singular)));
    setOkButtonText(tr("Exportieren"));

    setTargetFileBaseName(tr("%1liste").arg(db()->playersString(Database::Numerus::singular)));
    setTargetFileExtension(QString::fromUtf8("txt"));
    setTargetFileDescription(tr("Textdateien"));
    setFileDialogTitle(tr("Bitte einen Dateinamen für die %1liste auswählen").arg(
                          db()->playersString(Database::Numerus::singular)));

    // Players/Pairs selection

    QGroupBox *selectionBox = new QGroupBox(tr("%1auswahl").arg(
                                               db()->playersString(Database::Numerus::singular)));
    QVBoxLayout *selectionBoxLayout = new QVBoxLayout(selectionBox);
    layout()->addWidget(selectionBox);

    QRadioButton *exportAll = new QRadioButton(tr("Alle %1 exportieren (%2)").arg(
                                                  db()->playersString(Database::Numerus::plural),
                                                  QString::number(playersCount)));
    exportAll->setChecked(true);
    selectionBoxLayout->addWidget(exportAll);

    m_exportMarked = new QRadioButton(tr("Nur markierte %1 exportieren (%2)").arg(
                                         db()->playersString(Database::Numerus::plural),
                                         QString::number(markedPlayersCount)));
    m_exportMarked->setEnabled(markersSet && unMarkedPlayersThere);
    selectionBoxLayout->addWidget(m_exportMarked);

    m_exportUnMarked = new QRadioButton(tr("Nur unmarkierte %1 exportieren (%2)").arg(
                                           db()->playersString(Database::Numerus::plural),
                                           QString::number(playersCount - markedPlayersCount)));
    m_exportUnMarked->setEnabled(markersSet && unMarkedPlayersThere);
    connect(m_exportUnMarked, &QRadioButton::toggled,
            this, &ExportPlayersDialog::exportMarkedChanged);
    selectionBoxLayout->addWidget(m_exportUnMarked);

    // Export options

    QGroupBox *settingsBox = new QGroupBox(tr("Einstellungen"));
    layout()->addWidget(settingsBox);
    QVBoxLayout *settingsBoxLayout = new QVBoxLayout(settingsBox);

    m_exportMarkers = new QCheckBox(tr("Markierungen exportieren"));
    m_exportMarkers->setChecked(markersSet);
    m_exportMarkers->setEnabled(markersSet);
    settingsBoxLayout->addWidget(m_exportMarkers);

    m_openAfterExport = new QCheckBox(tr("Datei nach dem Export öffnen"));
    m_openAfterExport->setChecked(true);
    settingsBoxLayout->addWidget(m_openAfterExport);
}

void ExportPlayersDialog::exportMarkedChanged(bool checked)
{
    if (checked) {
        m_exportMakersState = m_exportMarkers->isChecked();
        m_exportMarkers->setChecked(false);
    } else {
        m_exportMarkers->setChecked(m_exportMakersState);
    }
    m_exportMarkers->setEnabled(! checked);
}

void ExportPlayersDialog::accept()
{
    QString fileName = getTargetFileName();
    if (fileName.isEmpty()) {
        return;
    }

    QFile playersListFile(fileName);
    playersListFile.open(QIODevice::WriteOnly | QIODevice::Text);

    QTextStream outStream(&playersListFile);
    outStream.setCodec(QTextCodec::codecForName("UTF-8"));

    emit updateStatus(tr("Exportiere %1liste …").arg(
                         db()->playersString(Database::Numerus::singular)));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    for (QString &pairOrPlayer : db()->listPlayers()) {
        if (m_exportMarked->isChecked()) {
            if (! pairOrPlayer.contains(MARKER_CHARACTER)) {
                continue;
            }
        } else if (m_exportUnMarked->isChecked()) {
            if (pairOrPlayer.contains(MARKER_CHARACTER)) {
                continue;
            }
        }

        if (m_exportMarkers->isChecked()) {
            outStream << pairOrPlayer + QString::fromUtf8("\n");
        } else {
            outStream << pairOrPlayer.remove(MARKER_CHARACTER) + QString::fromUtf8("\n");
        }
    }

    playersListFile.close();
    QApplication::restoreOverrideCursor();
    emit updateStatus(tr("%1liste in Datei „%2“ exportiert.").arg(
                         db()->playersString(Database::Numerus::singular),
                         QDir::toNativeSeparators(fileName)));

    successMessage(tr("%1listen-Export").arg(db()->playersString(Database::Numerus::singular)),
                   tr("Die %1liste wurde erfolgreich in die Datei „%2“ exportiert!").arg(
                      db()->playersString(Database::Numerus::singular),
                      QDir::toNativeSeparators(fileName)));

    if (m_openAfterExport->isChecked()) {
        openFile(fileName);
    }

    close();
}
