/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "MainWindow.h"
#include "Database.h"
#include "AboutDialog.h"
#include "PlayersPage.h"
#include "ScorePage.h"
#include "RankingPage.h"
#include "ExportTournamentDialog.h"
#include "StopWatch.h"
#include "TournamentInfoDialog.h"
#include "NewTournamentDialog.h"
#include "DefaultValues.h"
#include "ExportPlayersDialog.h"
#include "DefinedCharacters.h"
#include "BackupDialog.h"

// Qt includes
#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QTabWidget>
#include <QMenuBar>
#include <QAction>
#include <QApplication>
#include <QStatusBar>
#include <QMessageBox>
#include <QUrl>
#include <QDesktopServices>
#include <QFileInfo>
#include <QDebug>
#include <QFileDialog>
#include <QDir>
#include <QStandardPaths>
#include <QCoreApplication>
#include <QDirIterator>
#include <QDockWidget>
#include <QDesktopWidget>
#include <QPushButton>
#include <QTextCodec>
#include <QCloseEvent>
#include <QToolButton>
#include <QCheckBox>
#include <QRegularExpression>

MainWindow::MainWindow()
{
    // Setup the database
    m_db = new Database(this);
    connect(m_db, &Database::sqlError, this, &MainWindow::displaySqlError);

    // Window setup
    setWindowIcon(QIcon(resourcePath(QString::fromUtf8("muckturnier.svg"))));
    setWindowTitle(tr("Muckturnier"));

    QRect screenGeometry = QApplication::desktop()->screenGeometry();
    if (screenGeometry.width() >= 1024 && screenGeometry.height() >= 768) {
        resize(900, 700);
    } else {
        setWindowState(Qt::WindowMaximized);
    }

    m_centralWidget = new QWidget(this);
    setCentralWidget(m_centralWidget);
    QVBoxLayout *layout = new QVBoxLayout(m_centralWidget);

    // Banner

    QLabel *banner = new QLabel;
    banner->setPixmap(QIcon(resourcePath(QString::fromUtf8("banner.svg"))).pixmap(QSize(320, 80)));
    banner->setAlignment(Qt::AlignCenter);
    banner->setStyleSheet(QString::fromUtf8("background-color: #2f874c;"));
    banner->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    layout->addWidget(banner);
    layout->addStretch();

    // Startup buttons

    QHBoxLayout *buttonLayout = new QHBoxLayout;
    layout->addLayout(buttonLayout);

    QToolButton *newTournamentButton = new QToolButton;
    newTournamentButton->setAutoRaise(true);
    newTournamentButton->setStyleSheet(QString::fromUtf8("border: none;"));
    newTournamentButton->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    newTournamentButton->setIcon(QIcon(resourcePath(QString::fromUtf8("new.svg"))));
    newTournamentButton->setIconSize(QSize(100, 100));
    newTournamentButton->setText(tr("Ein neues Turnier starten"));
    connect(newTournamentButton, &QToolButton::clicked, this, &MainWindow::startNewTournament);
    buttonLayout->addWidget(newTournamentButton);

    QToolButton *loadTournamentButton = new QToolButton;
    loadTournamentButton->setAutoRaise(true);
    loadTournamentButton->setStyleSheet(QString::fromUtf8("border: none;"));
    loadTournamentButton->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    loadTournamentButton->setIcon(QIcon(resourcePath(QString::fromUtf8("load.svg"))));
    loadTournamentButton->setIconSize(QSize(100, 100));
    loadTournamentButton->setText(tr("Ein Turnier öffnen"));
    connect(loadTournamentButton, &QPushButton::clicked, this, &MainWindow::triggerLoadTournament);
    buttonLayout->addWidget(loadTournamentButton);

    // Footer

    QLabel *footer = new QLabel(tr("Das Open-Source-Projekt zum Auswerten von Muckturnieren"));
    footer->setAlignment(Qt::AlignCenter);
    footer->setStyleSheet(QString::fromUtf8("color: white; background-color: #309ac4;"));
    layout->addStretch();
    layout->addWidget(footer);

    // Dock setup

    setDockNestingEnabled(true);

    m_playersPage = new PlayersPage(this, m_db);
    connect(m_playersPage, &PlayersPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::tournamentCanBeStarted,
            this, &MainWindow::setScoreTabEnabled);
    m_playersDock = new QDockWidget(tr("Paare"));
    m_playersDock->setWidget(m_playersPage);
    formatDockWidget(m_playersDock);
    m_playersDock->hide();

    m_scorePage = new ScorePage(this, m_db);
    connect(m_scorePage, &ScorePage::tournamentStarted, this, &MainWindow::tournamentStarted);
    connect(m_scorePage, &ScorePage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_scorePage, &ScorePage::scoreChanged, m_playersPage, &PlayersPage::scoreChanged);
    m_scoreDock = new QDockWidget(tr("Ergebnisse"));
    m_scoreDock->setWidget(m_scorePage);
    formatDockWidget(m_scoreDock);
    tabifyDockWidget(m_playersDock, m_scoreDock);
    m_scoreDock->hide();

    m_rankingPage = new RankingPage(this, m_db);
    connect(m_rankingPage, &RankingPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::playersListChanged,
            m_rankingPage, &RankingPage::dataChanged);
    connect(m_scorePage, &ScorePage::scoreChanged, m_rankingPage, &RankingPage::dataChanged);
    m_rankingDock = new QDockWidget(tr("Rangliste"));
    m_rankingDock->setWidget(m_rankingPage);
    formatDockWidget(m_rankingDock);
    tabifyDockWidget(m_scoreDock, m_rankingDock);
    m_rankingDock->hide();

    // Menu

    // "File"

    QMenu *fileMenu = menuBar()->addMenu(tr("Datei"));

    QAction *newTournamentAction = new QAction(tr("Neues Turnier starten"), this);
    fileMenu->addAction(newTournamentAction);
    connect(newTournamentAction, &QAction::triggered, this, &MainWindow::startNewTournament);

    QAction *loadTournamentAction = new QAction(tr("Turnier öffnen"), this);
    fileMenu->addAction(loadTournamentAction);
    connect(loadTournamentAction, &QAction::triggered, this, &MainWindow::triggerLoadTournament);

    QMenu *imExportMenu = fileMenu->addMenu(tr("Import und Export"));

    m_importPlayersListAction = new QAction(tr("Paarliste importieren"), this);
    connect(m_importPlayersListAction, &QAction::triggered, this, &MainWindow::importPlayersList);
    imExportMenu->addAction(m_importPlayersListAction);
    m_importPlayersListAction->setEnabled(false);

    m_exportPlayersListAction = new QAction(tr("Paarliste exportieren"), this);
    connect(m_playersPage, &PlayersPage::enableExport,
            m_exportPlayersListAction, &QAction::setEnabled);
    connect(m_exportPlayersListAction, &QAction::triggered, this, &MainWindow::exportPlayersList);
    imExportMenu->addAction(m_exportPlayersListAction);
    m_exportPlayersListAction->setEnabled(false);

    imExportMenu->addSeparator();

    m_exportTournamentAction = new QAction(tr("Turnierdaten exportieren"), this);
    connect(m_rankingPage, &RankingPage::enableExport,
            m_exportTournamentAction, &QAction::setEnabled);
    connect(m_exportTournamentAction, &QAction::triggered,
            this, &MainWindow::showExportTournamentDialog);
    imExportMenu->addAction(m_exportTournamentAction);
    m_exportTournamentAction->setEnabled(false);

    fileMenu->addSeparator();

    m_tournamentInfoAction = new QAction(tr("Turnierinformationen"), this);
    connect(m_tournamentInfoAction, &QAction::triggered, this, &MainWindow::showTournamentInfo);
    fileMenu->addAction(m_tournamentInfoAction);
    m_tournamentInfoAction->setEnabled(false);

    fileMenu->addSeparator();

    m_closeTournamentAction = new QAction(tr("Turnier schließen"), this);
    fileMenu->addAction(m_closeTournamentAction);
    m_closeTournamentAction->setEnabled(false);
    connect(m_closeTournamentAction, &QAction::triggered, this, &MainWindow::closeTournament);

    QAction *quitAction = new QAction(tr("Beenden"), this);
    fileMenu->addAction(quitAction);
    connect(quitAction, &QAction::triggered, this, &QWidget::close);

    // "View"

    QMenu *viewMenu = menuBar()->addMenu(tr("Ansicht"));

    QAction *arrangeTabifiedAction = new QAction(tr("Seiten als Tabs"), this);
    connect(arrangeTabifiedAction, &QAction::triggered, this, &MainWindow::arrangeTabified);
    viewMenu->addAction(arrangeTabifiedAction);

    QAction *arrangeHorizontallyAction = new QAction(tr("Seiten nebeneinander"), this);
    connect(arrangeHorizontallyAction, &QAction::triggered, this, &MainWindow::arrangeHorizontally);
    viewMenu->addAction(arrangeHorizontallyAction);

    QAction *arrangeVerticallyAction = new QAction(tr("Seiten übereinander"), this);
    connect(arrangeVerticallyAction, &QAction::triggered, this, &MainWindow::arrangeVertically);
    viewMenu->addAction(arrangeVerticallyAction);

    QAction *arrangeOwnWindowAction = new QAction(tr("Seiten als einzelne Fenster"), this);
    connect(arrangeOwnWindowAction, &QAction::triggered, this, &MainWindow::arrangeOwnWindow);
    viewMenu->addAction(arrangeOwnWindowAction);

    // "Extras"

    QMenu *extrasMenu = menuBar()->addMenu(tr("Extras"));

    QAction *stopWatchAction = new QAction(tr("Stoppuhr"), this);
    extrasMenu->addAction(stopWatchAction);
    connect(stopWatchAction, &QAction::triggered, this, &MainWindow::showStopWatch);

    extrasMenu->addSeparator();

    m_resetDatabaseAction = new QAction(tr("Datenbank zurücksetzen"), this);
    connect(m_playersPage, &PlayersPage::enableExport,
            m_resetDatabaseAction, &QAction::setEnabled);
    extrasMenu->addAction(m_resetDatabaseAction);
    connect(m_resetDatabaseAction, &QAction::triggered, this, &MainWindow::resetDatabase);
    m_resetDatabaseAction->setEnabled(false);

    extrasMenu->addSeparator();

    m_createDatabaseBackupAction = new QAction(tr("Backup der Datenbank anlegen"), this);
    connect(m_createDatabaseBackupAction, &QAction::triggered,
            this, &MainWindow::createDatabaseBackup);
    extrasMenu->addAction(m_createDatabaseBackupAction);
    m_createDatabaseBackupAction->setEnabled(false);

    m_restoreDatabaseBackupAction = new QAction(tr("Backup wiederherstellen"), this);
    connect(m_restoreDatabaseBackupAction, &QAction::triggered,
            this, &MainWindow::restoreDatabaseBackup);
    extrasMenu->addAction(m_restoreDatabaseBackupAction);
    m_restoreDatabaseBackupAction->setEnabled(false);

    // "Help"

    QMenu *helpMenu = menuBar()->addMenu(tr("Hilfe"));

    QAction *handbookAction = new QAction(tr("Handbuch zu Muckturnier"), this);
    helpMenu->addAction(handbookAction);
    connect(handbookAction, &QAction::triggered, this, &MainWindow::showHandbook);

    QAction *aboutAction = new QAction(tr("Über Muckturnier"), this);
    helpMenu->addAction(aboutAction);
    connect(aboutAction, &QAction::triggered, this, &MainWindow::showAboutDialog);

    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::startNewTournament()
{
    NewTournamentDialog *dialog = new NewTournamentDialog(this, m_db);
    connect(dialog, &NewTournamentDialog::updateStatus, this, &MainWindow::updateStatus);
    if (! dialog->exec()) {
        return;
    }

    dbReset();
    databaseChanged();
}

void MainWindow::triggerLoadTournament()
{
    loadTournament();
}

void MainWindow::loadTournament(QString dbFileName)
{
    if (dbFileName.isEmpty()) {
        dbFileName = QFileDialog::getOpenFileName(this,
            tr("Bitte eine Turnierdatenbank auswählen"),
            QDir::toNativeSeparators(QStandardPaths::writableLocation(
                QStandardPaths::DocumentsLocation)),
            tr("Muckturnier-Datenbanken") + QString::fromUtf8(" (*.mtdb *.db);;")
            + tr("Alle Dateien") + QString::fromUtf8(" (*)"));
        if (dbFileName.isEmpty()) {
            return;
        }
    }

    updateStatus(tr("Öffne Turnierdatenbank …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);
    Database::Error openResult = m_db->openDb(dbFileName);
    QApplication::restoreOverrideCursor();

    switch (openResult) {
    case Database::Error::NoError:
        break;
    case Database::Error::NoErrorReadOnly:
        QMessageBox::warning(this,
            tr("Turnier öffnen"),
            tr("Die Datei „%1“ ist nicht beschreibbar. Die Turnierdatenbank kann nur angezeigt, "
               "aber nicht verändert werden.").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        break;
    case Database::Error::DbVersion3:
        updateDatabase(dbFileName, 3);
        return;
    case Database::Error::DbVersion4:
        updateDatabase(dbFileName, 4);
        return;
    case Database::Error::DbVersion5:
        updateDatabase(dbFileName, 5);
        return;
    case Database::Error::CouldNotOpen:
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("<p><b>Das Öffnen der Turnierdatenbank ist nicht möglich:</b></p>")
            + m_db->errorMessage(),
            QMessageBox::Ok);
        return;
    }

    databaseChanged();

    if (m_db->tournamentStarted()) {
        m_scoreDock->raise();
    } else {
        // The score page could be enabled by the previously used database,
        // so disable it first
        m_scoreDock->hide();
        // Re-enable the score page if the database already contains
        // the necessary number of players/pairs
        m_playersPage->reload();
        m_playersDock->raise();
    }

    m_tournamentInfoAction->setEnabled(true);
    m_createDatabaseBackupAction->setEnabled(true);
    m_restoreDatabaseBackupAction->setEnabled(m_db->isWritable());

    updateStatus(tr("Turnierdatenbank %1 geöffnet").arg(QDir::toNativeSeparators(dbFileName)));
    QApplication::restoreOverrideCursor();
}

void MainWindow::updateDatabase(const QString &dbFileName, int dbVersion)
{
    QFileInfo fileInfo(dbFileName);

    if (! fileInfo.isWritable()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("<p>Das Datenbankformat von Muckturnier hat sich geändert. Die Turnierdatenbank "
               "„%1“ hat die Revision %2, aktuell ist Revision %3.</p>"
               "<p>Um die Datei benutzen zu können, müsste ein Update der Datenbank durchgeführt "
               "werden, aber die Datei ist nicht beschreibbar.</p>"
               "<p>Bitte die Zugriffsrechte überprüfen, und die Datei erneut laden!</p>").arg(
               QDir::toNativeSeparators(dbFileName), QString::number(dbVersion),
                                                     QString::number(m_db->getDbVersion())),
            QMessageBox::Ok);
        return;
    }

    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("Update des Datenbankformats"));
    messageBox.setText(tr("<p><b>Das Datenbankformat von Muckturnier hat sich geändert.</b></p>"
                          "<p>Soll die Turnierdatenbank „%1“ aktualisiert werden? Es wird dabei "
                          "ein Backup der Original-Datenbank im selben Ordner angelegt.</p>").arg(
                          QDir::toNativeSeparators(dbFileName)));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::Cancel);
    messageBox.setDefaultButton(QMessageBox::Yes);

    if (messageBox.exec() != QMessageBox::Yes) {
        updateStatus(tr("Datenbankupdate abgebrochen!"));
        QMessageBox::critical(this,
            tr("Update des Datenbankformats"),
            tr("Das Update des Datenbankformats wurde abgebrochen. Die Datenbank „%1“ kann nicht "
               "geöffnet werden.").arg(QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Datenbankupdate läuft …"));

    m_db->closeDatabaseConnection();
    QString backupFileName = BackupDialog::createBackup(dbFileName, QString(),
                                                        QString::fromUtf8("original_dbv")
                                                        + QString::number(dbVersion));
    if (backupFileName.isEmpty()) {
        QApplication::restoreOverrideCursor();
        QMessageBox::critical(this,
            tr("Update des Datenbankformats"),
            tr("<p><b>Die Backup-Datei konnte nicht angelegt werden.</b></p>"
               "<p>Das Update wird abgebrochen. Bitte die Zugriffsrechte (des Verzeichnisses „%2“) "
               "überprüfen!</p>").arg(
               QDir::toNativeSeparators(fileInfo.path())),
            QMessageBox::Ok);
        updateStatus(tr("Datenbankupdate abgebrochen!"));
        return;
    }

    if (dbVersion == 3) {
        m_db->updateDatabase3to4(dbFileName);
        m_db->updateDatabase4to5(dbFileName);
        m_db->updateDatabase5to6(dbFileName);
    } else if (dbVersion == 4) {
        m_db->updateDatabase4to5(dbFileName);
        m_db->updateDatabase5to6(dbFileName);
    } else if (dbVersion == 5) {
        m_db->updateDatabase5to6(dbFileName);
    }

    QApplication::restoreOverrideCursor();
    updateStatus(tr("Datenbankupdate beendet!"));
    QMessageBox::information(this,
        tr("Update des Datenbankformats"),
        tr("<p><b>Die Datenbank „%1“ wurde erfolgreich aktualisiert!</b></p>"
           "<p>Ein Backup der Originaldatei wurde mit dem Namen „%2“ erstellt.</p>").arg(
           QDir::toNativeSeparators(dbFileName),
           QDir::toNativeSeparators(backupFileName)),
        QMessageBox::Ok);

    loadTournament(dbFileName);
}

void MainWindow::showHandbook()
{
    // Opens the default browser to display the handbook.

    QString readmePath = resourcePath(QString::fromUtf8("Readme.htm"));
    if (! readmePath.isEmpty()) {
        QDesktopServices::openUrl(QUrl::fromLocalFile(readmePath));
        return;
    }

    QMessageBox::critical(this,
                          tr("Handbuch anzeigen"),
                          tr("<p>Das Handbuch kann nicht angezeigt werden: Die Datei „Readme.htm“ "
                             "wurde nicht gefunden.</p>"
                             "<p>Online ist unter <a href=\"http://muckturnier.org/readme/\">"
                             "http://muckturnier.org/readme/</a> das aktuelle Handbuch zu finden."
                             "</p>"),
                          QMessageBox::Ok);
}

void MainWindow::showAboutDialog()
{
    QDialog *aboutDialog = new AboutDialog(this, m_db,
                                           resourcePath(QString::fromUtf8("banner.svg")));
    aboutDialog->show();
}

void MainWindow::updateStatus(const QString &message)
{
    statusBar()->showMessage(message, 5000);
}

void MainWindow::databaseChanged()
{
    setWindowTitle(tr("Muckturnier: %1").arg(QFileInfo(m_db->getDbFile()).fileName()));
    m_centralWidget->hide();
    setPlayersTabName();
    m_playersDock->show();
    m_playersPage->reload();

    m_exportTournamentAction->setEnabled(false);
    m_closeTournamentAction->setEnabled(true);
    m_tournamentInfoAction->setEnabled(true);
    m_createDatabaseBackupAction->setEnabled(true);
    m_restoreDatabaseBackupAction->setEnabled(m_db->isWritable());

    checkEnableRankingPage();
    if (m_db->tournamentStarted()) {
        m_rankingPage->dataChanged();
        m_importPlayersListAction->setEnabled(false);
    } else {
        m_importPlayersListAction->setEnabled(m_db->isWritable());
    }

    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        m_exportPlayersListAction->setText(tr("Paarliste exportieren"));
        m_importPlayersListAction->setText(tr("Paarliste importieren"));
        m_importPlayersListString = tr("Paarliste");
        // Extra variable because the action will get a shortcut ampersand somewhere
        m_importPlayersListActionName = tr("Paarliste importieren");
        break;
    case Database::TournamentMode::singlePlayers:
        m_exportPlayersListAction->setText(tr("Spielerliste exportieren"));
        m_importPlayersListAction->setText(tr("Spielerliste importieren"));
        m_importPlayersListString = tr("Spielerliste");
        m_importPlayersListActionName = tr("Spielerliste importieren");
        break;
    }

    m_resetDatabaseAction->setEnabled(m_db->isWritable());

    // Restore the state of the options we save in the database
    m_scorePage->setNoTableNumbers(m_db->noTableNumbers());
    m_scorePage->setAutoSelectPairs(m_db->autoSelectPairs());
    m_rankingPage->setIncludeOpponentGoals(m_db->includeOpponentGoals());

    // Be sure to have the correct pairs/players selected
    m_scorePage->checkCorrectPairSelection();
}

void MainWindow::settingsSaved()
{
    setPlayersTabName();
    m_playersDock->show();
    m_playersPage->reload();
    m_playersDock->raise();
    m_tournamentInfoAction->setEnabled(true);
    m_createDatabaseBackupAction->setEnabled(true);
    m_restoreDatabaseBackupAction->setEnabled(m_db->isWritable());
}

void MainWindow::setScoreTabEnabled(bool state)
{
    m_scoreDock->setVisible(state);
    if(state == true) {
        m_scorePage->reload();
    }
}

void MainWindow::tournamentStarted()
{
    setPlayersTabName();
    m_playersPage->reload();
    checkEnableRankingPage();
    m_importPlayersListAction->setEnabled(false);
}

void MainWindow::setPlayersTabName()
{
    m_playersDock->setWindowTitle(m_db->playersString(Database::Numerus::plural));
}

void MainWindow::checkEnableRankingPage()
{
    m_rankingDock->setVisible(m_db->tournamentStarted());
}

void MainWindow::resizeEvent(QResizeEvent *event)
{
   QMainWindow::resizeEvent(event);
   if (m_scoreDock->isVisible()) {
        m_scorePage->setScrollAreaHeight();
   }
}

void MainWindow::showExportTournamentDialog()
{
    ExportTournamentDialog *exportTournamentDialog = new ExportTournamentDialog(this, m_db,
                                                                                m_rankingPage);
    connect(exportTournamentDialog, &ExportTournamentDialog::updateStatus,
            this, &MainWindow::updateStatus);
    exportTournamentDialog->exec();
}

void MainWindow::showStopWatch()
{
    StopWatch *stopWatch = new StopWatch(this);
    stopWatch->show();
}

void MainWindow::showTournamentInfo()
{
    TournamentInfoDialog *tournamentInfoDialog = new TournamentInfoDialog(this, m_db);
    tournamentInfoDialog->exec();
}

void MainWindow::formatDockWidget(QDockWidget *dock)
{
    dock->setContextMenuPolicy(Qt::PreventContextMenu);
    dock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);
    dock->setStyleSheet(QString::fromUtf8("QDockWidget { font-weight: bold; font-size: %1pt }").arg(
        (int) double(dock->font().pointSize()) * 1.2));
    addDockWidget(Qt::TopDockWidgetArea, dock);
}

QString MainWindow::resourcePath(const QString &fileName)
{
    if (m_possibleDataLocations.size() == 0) {
        // Get a list of possible locations for our external files
        QStringList possibleDataLocations;
        const QString iconsPath         = QString::fromUtf8("icons/hicolor/scalable/apps/");
        const QString resourcesPath     = QString::fromUtf8("muckturnier/");
        const QString documentationPath = QString::fromUtf8("doc/muckturnier/");

        // Add the binary's directory (possibly a build directory)
        possibleDataLocations << QCoreApplication::applicationDirPath() + QString::fromUtf8("/");

        // Add a possibly prefixed directory tree
        QUrl prefixed = QUrl(QCoreApplication::applicationDirPath() + QString::fromUtf8("/"))
                            .resolved(QUrl(QString::fromUtf8("../share/")));
        possibleDataLocations << prefixed.toString() + iconsPath
                              << prefixed.toString() + resourcesPath
                              << prefixed.toString() + documentationPath;

        // Add system-wide locations
        for (const QString &path
            : QStandardPaths::standardLocations(QStandardPaths::GenericDataLocation)) {
            possibleDataLocations << path + QString::fromUtf8("/") + iconsPath
                                  << path + QString::fromUtf8("/") + resourcesPath
                                  << path + QString::fromUtf8("/") + documentationPath;
        }

        // Reverse the list
        while (possibleDataLocations.size() > 0) {
            const QString path = QDir::toNativeSeparators(possibleDataLocations.takeLast());
            if (! m_possibleDataLocations.contains(path)) {
                m_possibleDataLocations << path;
            }
        }
    }

    // Check if we already have it
    if (m_resourcePath.contains(fileName)) {
        return m_resourcePath[fileName];
    }

    // First, we search for embedded resources ...
    QDirIterator it(QString::fromUtf8(":"), QDirIterator::Subdirectories);
    while (it.hasNext()) {
        if (it.next() == QString::fromUtf8(":/") + fileName) {
            m_resourcePath[fileName] = QString::fromUtf8(":/") + fileName;
            return m_resourcePath[fileName];
        }
    }

    // ... then we check all other possible data locations
    for (const QString &path : m_possibleDataLocations) {
        QString filePath = path + fileName;
        if (QFileInfo::exists(filePath)) {
            m_resourcePath[fileName] = filePath;
            return m_resourcePath[fileName];
        }
    }

    // This will also return QString() if nothing is found
    return m_resourcePath[fileName];
}

void MainWindow::closeTournament()
{
    m_db->closeDatabaseConnection();
    m_playersDock->hide();
    m_scoreDock->hide();
    m_rankingDock->hide();
    m_centralWidget->show();
    m_closeTournamentAction->setEnabled(false);
    m_tournamentInfoAction->setEnabled(false);
    m_createDatabaseBackupAction->setEnabled(false);
    m_restoreDatabaseBackupAction->setEnabled(false);
    m_resetDatabaseAction->setEnabled(false);
    m_exportPlayersListAction->setEnabled(false);
    m_exportTournamentAction->setEnabled(false);
    m_importPlayersListAction->setEnabled(false);
    setWindowTitle(tr("Muckturnier"));
    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::exportPlayersList()
{
    ExportPlayersDialog *dialog = new ExportPlayersDialog(this, m_db);
    connect(dialog, &ExportPlayersDialog::updateStatus, this, &MainWindow::updateStatus);
    dialog->exec();
    return;
}

void MainWindow::attachDocks(Qt::DockWidgetArea area)
{
    m_playersDock->setFloating(false);
    m_scoreDock->setFloating(false);
    m_rankingDock->setFloating(false);
    addDockWidget(area, m_playersDock);
    addDockWidget(area, m_scoreDock);
    addDockWidget(area, m_rankingDock);
}

void MainWindow::arrangeTabified()
{
    attachDocks(Qt::TopDockWidgetArea);
    tabifyDockWidget(m_playersDock, m_scoreDock);
    tabifyDockWidget(m_scoreDock, m_rankingDock);
}

void MainWindow::arrangeHorizontally()
{
    attachDocks(Qt::TopDockWidgetArea);
    splitDockWidget(m_playersDock, m_scoreDock, Qt::Horizontal);
    splitDockWidget(m_scoreDock, m_rankingDock, Qt::Horizontal);
}

void MainWindow::arrangeVertically()
{
    attachDocks(Qt::LeftDockWidgetArea);
    splitDockWidget(m_playersDock, m_scoreDock, Qt::Vertical);
    splitDockWidget(m_scoreDock, m_rankingDock, Qt::Vertical);
}

void MainWindow::arrangeOwnWindow()
{
    m_playersDock->setFloating(true);
    m_scoreDock->setFloating(true);
    m_rankingDock->setFloating(true);
}

void MainWindow::dbReset()
{
    setPlayersTabName();
    m_playersDock->hide();
    m_playersPage->reset();
    m_scoreDock->hide();
    m_rankingDock->hide();
    m_importPlayersListAction->setEnabled(true);
    m_exportPlayersListAction->setEnabled(false);
    m_exportTournamentAction->setEnabled(false);
}

void MainWindow::resetDatabase()
{
    QCheckBox *createBackup = new QCheckBox(tr("Vorher ein Backup der Datenbank anlegen"));
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Warning);
    messageBox.setWindowTitle(tr("Datenbank zurücksetzen"));
    messageBox.setText(tr("Sollen wirklich alle angemeldeten Spieler und alle eingegebenen "
                          "Ergebnisse gelöscht werden?"));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::Cancel);
    messageBox.setDefaultButton(QMessageBox::Cancel);
    messageBox.setCheckBox(createBackup);

    if (messageBox.exec() == QMessageBox::Cancel) {
        return;
    }

    if (createBackup->isChecked()) {
        if (! createDatabaseBackup()) {
            QMessageBox::warning(this,
                tr("Datenbank zurücksetzen"),
                tr("Das angefragte Datenbankbackup konnte nicht angelegt werden, deswegen wurde "
                   "das Zurücksetzen der Datenbank abgebrochen."),
                QMessageBox::Ok);
            return;
        }
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Setzte Turnierdatenbank zurück …"));

    m_db->reset();
    setPlayersTabName();
    dbReset();
    settingsSaved();

    QApplication::restoreOverrideCursor();
    updateStatus(tr("Turnierdatenbank zurückgesetzt."));

    QMessageBox::information(this,
        tr("Datenbank zurücksetzen"),
        tr("Die Datenbank wurde zurückgesetzt."),
        QMessageBox::Ok
    );
}

void MainWindow::importPlayersList()
{
    // Get the import file's name
    QString inputFileName = QFileDialog::getOpenFileName(this,
        tr("Bitte eine %1 auswählen").arg(m_importPlayersListString),
        QDir::toNativeSeparators(
            QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)),
        tr("Textdateien") + QString::fromUtf8(" (*.txt);;")
        + tr("Alle Dateien") + QString::fromUtf8(" (*)"));
    if (inputFileName.isEmpty()) {
        return;
    }

    // Try to open it
    QFile inputFile(inputFileName);
    if (! inputFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QMessageBox::critical(this,
            m_importPlayersListActionName,
            tr("Es bestehen keine Leserechte für die Datei „%1“. Bitte die Zugriffsrechte "
               "überprüfen!").arg(QDir::toNativeSeparators(inputFileName)),
            QMessageBox::Ok);
        return;
    }

    // Check if it's suspiciously large (more than about 200 mean pair names)
    QFileInfo fileInfo(inputFileName);
    if (fileInfo.size() > IMPORT_SIZE_WARNING) {
        QMessageBox messageBox(this);
        messageBox.setIcon(QMessageBox::Warning);
        messageBox.setWindowTitle(m_importPlayersListActionName);
        messageBox.setText(tr("Die Datei „%1“ ist ungewöhnlich groß für eine %1. Soll sie trotzdem "
                              "importiert werden?").arg(QDir::toNativeSeparators(inputFileName),
                                                        m_importPlayersListString));
        messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        messageBox.setDefaultButton(QMessageBox::No);

        if (messageBox.exec() == QMessageBox::No) {
            return;
        }
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Importiere %1 …").arg(m_importPlayersListString));

    bool markersPresent = false;
    QVariantList playersToAdd;
    int importedPlayers = 0;
    int lineNumber = 0;

    // Get the current player list to search for duplicates
    QStringList allPlayers = m_db->listPlayers();
    for (QString &name : allPlayers) {
        name.remove(MARKER_CHARACTER);
        name = name.toLower();
    }

    // Read the file line per line

    QTextStream stream(&inputFile);
    stream.setCodec(QTextCodec::codecForName("UTF-8"));

    while (! stream.atEnd()) {
        lineNumber++;
        // Let's truncate lines at byte 256
        // just in case some user opens some binary file or whatnot ;-)
        QString playersName = stream.readLine(256).simplified();
        if (playersName.isEmpty()) {
            continue;
        }

        // If the file contains invalid characters, refuse to import it.
        if (playersName.contains(REPLACEMENT_CHARACTER)) {
            QString errorLine;
            if (playersName.length() > 45) {
                errorLine = playersName.left(45) + QString::fromUtf8("…");
            } else {
                errorLine = playersName;
            }

            QApplication::restoreOverrideCursor();
            QMessageBox::critical(this,
                m_importPlayersListActionName,
                tr("<p>Die Datei „%1“ enthält ungültige Zeichen. Vermutlich ist es keine reine "
                   "Textdatei, oder sie ist nicht UTF-8-kodiert. Zeile Nr. %2 war die erste "
                   "fehlerhalt kodierte:</p><p>%3</p>"
                   "<p>Es kann eine UTF-8-kodierte Textdatei importiert werden, die pro Zeile "
                   "einen Paar- bzw. Spielernamen enthält.</p>").arg(
                   QDir::toNativeSeparators(inputFileName), QString::number(lineNumber), errorLine),
                   QMessageBox::Ok);
            return;
        }

        // Check for suspiciously long lines
        if (playersName.length() > IMPORT_LINE_LENGTH_WARNING) {
            QApplication::restoreOverrideCursor();
            QMessageBox messageBox(this);
            messageBox.setIcon(QMessageBox::Warning);
            messageBox.setWindowTitle(m_importPlayersListActionName);
            messageBox.setText(tr("Die Datei „%1“ enthält ungewöhnlich lange Zeilen. Soll sie "
                                  "trotzdem importiert werden?").arg(
                                  QDir::toNativeSeparators(inputFileName)));
            messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
            messageBox.setDefaultButton(QMessageBox::No);

            if (messageBox.exec() == QMessageBox::No) {
                return;
            }
        }

        // Check for a duplicate existing name and generate non-colliding names.

        int i = 1;
        bool hasMark = false;
        bool hasTopMark = false;

        if (playersName.startsWith(MARKER_CHARACTER)) {
            hasTopMark = true;
        } else if (playersName.endsWith(MARKER_CHARACTER)) {
            hasMark = true;
        }
        playersName.remove(MARKER_CHARACTER);

        QString checkName = playersName.toLower();
        while (true) {
            bool found = false;
            for (const QString &name : allPlayers) {
                if (name == checkName) {
                    found = true;
                    break;
                }
            }
            if (! found) {
                break;
            } else {
                checkName = playersName.toLower() + QString::fromUtf8(" (%1)").arg(++i);
            }
        }
        if (i > 1) {
            playersName = playersName + QString::fromUtf8(" (%1)").arg(i);
        }

        allPlayers << playersName.toLower();
        if (hasTopMark) {
            playersName = MARKER_CHARACTER + playersName;
        } else if (hasMark) {
            playersName += MARKER_CHARACTER;
        }
        playersToAdd << playersName;
        importedPlayers++;

        if (! markersPresent) {
            if (playersName.contains(MARKER_CHARACTER)) {
                markersPresent = true;
            }
        }
    }

    inputFile.close();

    // Check for suspiciously many pairs/players
    if (importedPlayers > IMPORT_LINES_COUNT_WARNING) {
        QApplication::restoreOverrideCursor();
        QMessageBox messageBox(this);
        messageBox.setIcon(QMessageBox::Warning);
        messageBox.setWindowTitle(m_importPlayersListActionName);
        messageBox.setText(tr("Die Datei „%1“ enthält %2 Zeilen! Soll sie wirklich importiert "
                              "werden?").arg(QDir::toNativeSeparators(inputFileName),
                                             QString::number(importedPlayers)));
        messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        messageBox.setDefaultButton(QMessageBox::No);

        if (messageBox.exec() == QMessageBox::No) {
            return;
        }
    }

    // Check if we have markers and want to import them
    if (markersPresent) {
        QApplication::restoreOverrideCursor();
        QMessageBox messageBox(this);
        messageBox.setIcon(QMessageBox::Question);
        messageBox.setWindowTitle(m_importPlayersListActionName);
        messageBox.setText(tr("Die Datei „%1“ enthält markierte Spielernamen. Sollen die "
                              "Markierungen importiert werden?").arg(
                              QDir::toNativeSeparators(inputFileName)));
        messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        messageBox.setDefaultButton(QMessageBox::Yes);

        if (messageBox.exec() == QMessageBox::No) {
            for (QVariant &name : playersToAdd) {
                QString nameString = name.toString();
                if (nameString.contains(MARKER_CHARACTER)) {
                    nameString.remove(MARKER_CHARACTER);
                    name = QVariant(nameString);
                }
            }
        }
    }

    // Could be restored by a previous dialog
    if (! QGuiApplication::overrideCursor()) {
        QApplication::setOverrideCursor(Qt::WaitCursor);
    }

    // Do the actual import
    if (importedPlayers > 0) {
        m_db->registerPlayers(playersToAdd);
        m_playersPage->reload();
        m_exportPlayersListAction->setEnabled(true);
        m_resetDatabaseAction->setEnabled(true);
    }

    QApplication::restoreOverrideCursor();
    updateStatus(tr("%1 importiert.").arg(m_importPlayersListString));
    QMessageBox::information(this,
        m_importPlayersListActionName,
        tr("Es wurden %1 %2 importiert!").arg(
            QString::number(importedPlayers),
            m_db->playersString((importedPlayers == 1) ? Database::Numerus::singular
                                                       : Database::Numerus::plural)),
        QMessageBox::Ok);
}

void MainWindow::displaySqlError(const QString &error, const QString &query)
{
    // This should not happen

    if (QApplication::overrideCursor() != nullptr) {
        QApplication::restoreOverrideCursor();
    }

    QString message = tr("<p>Es ist ein unerwarteter Fehler bei einer Datenbankabfrage "
                         "aufgetreten. Das sollte nicht passieren!</p>"
                         "<p>Evtl. wurde die Datenbank gelöscht oder verschoben, oder die "
                         "Zugriffsrechte wurden nach dem Öffnen verändert?</p>");
    if (! query.isEmpty()) {
        message.append(tr("<p><b>Die Anfrage war:</b></p><p>%1</p>").arg(query));
    }

    message.append(tr("<p><b>Die Fehlermeldung war:</b></p>%1").arg(error));

    QMessageBox::critical(this,
        tr("Fehler in der Datenbankkommunikation"),
        message,
        QMessageBox::Ok
    );
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    bool stopWatchRunning = false;;
    for (QWidget *stopWatch : findChildren<StopWatch *>()) {
        if (dynamic_cast<StopWatch *>(stopWatch)->isRunning()) {
            stopWatchRunning = true;
            break;
        }
    }

    if (! stopWatchRunning) {
        event->accept();
        return;
    }

    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Warning);
    messageBox.setWindowTitle(tr("Muckturnier beenden"));
    messageBox.setText(tr("<p>Es läuft mindestens eine Stoppuhr! Alle Stoppuhren werden beim "
                          "Schließen ebenfalls geschlossen, und die Zeitinformation geht "
                          "verloren!</p>"
                          "<p>Soll Muckturnier trotzdem geschlossen werden?</p>"));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        event->ignore();
    } else {
        event->accept();
    }
}

bool MainWindow::createDatabaseBackup()
{
    QString dbFileName = m_db->getDbFile();
    BackupDialog::DbFileInfo info = BackupDialog::parseDbFileName(dbFileName);
    QString baseName;

    // Let's see if a backup file seems to be open
    if (! info.originalBaseName.isEmpty()) {
        QMessageBox messageBox(this);
        messageBox.setWindowTitle(tr("Backup der Datenbank anlegen"));
        messageBox.setText(tr("<p><b>Ist ein Backup offen?</b></p>"
                              "<p>Es sieht so aus, als wäre die offene Turnierdatenbank „%1“ ein "
                              "Backup von „%2.mtdb“ (erstellt %3).</p>"
                              "<p>Soll im Dateinamen des jetzt angefragten Backups das Datum und "
                              "die Zeit aktualisiert oder zusätzlich angehängt werden?</p>").arg(
                              QDir::toNativeSeparators(dbFileName),
                              QDir::toNativeSeparators(info.originalBaseName),
                              BackupDialog::getDateString(info.backupDate)));
        QAbstractButton *updateButton = messageBox.addButton(tr("Aktualisieren"),
                                                             QMessageBox::AcceptRole);
        messageBox.addButton(tr("Anhängen"), QMessageBox::RejectRole);
        QPushButton *cancelButton = messageBox.addButton(QMessageBox::Cancel);
        messageBox.exec();
        if (messageBox.clickedButton() == cancelButton) {
           return false;
        }
        if (messageBox.clickedButton() == updateButton) {
            baseName = info.path + QString::fromUtf8("/") + info.originalBaseName;
        }
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Backup wird angelegt …"));
    m_db->closeDatabaseConnection();
    const QString backupFileName = BackupDialog::createDatabaseBackup(dbFileName, baseName);
    bool backupFailed = backupFileName.isEmpty();
    QApplication::restoreOverrideCursor();

    if (backupFailed) {
        updateStatus(tr("Backup konnte nicht erstellt werden!"));
        QMessageBox::critical(this,
            tr("Backup der Datenbank anlegen"),
            tr("<p><b>Das Anlegen eines Backups ist fehlgeschlagen!</b></p>"
               "<p>Die Datei konnte nicht angelegt werden. Bitte die Zugriffsrechte von „%1“ "
               "überprüfen!</p>").arg(
               QDir::toNativeSeparators(info.path)
            ),
            QMessageBox::Ok);
    } else {
        updateStatus(tr("Backup erfolgreich erstellt!"));
        QMessageBox::information(this,
            tr("Backup der Datenbank anlegen"),
            tr("Es wurde erfolgreich ein Backup der Datenbank mit dem Dateinamen „%1“ "
               "erstellt.").arg(
               QDir::toNativeSeparators(backupFileName)
            ),
            QMessageBox::Ok);
    }

    m_db->openDb(dbFileName);
    return backupFailed;
}

void MainWindow::restoreDatabaseBackup()
{
    BackupDialog dialog(this, m_db->getDbFile());
    if (! dialog.exec()) {
        return;
    }

    QFileInfo backupInfo(dialog.backupToRestore());
    QFileInfo dbToOverwriteInfo(dialog.dbToOverwrite());

    if (QMessageBox::question(this,
            tr("Backup wiederherstellen"),
            tr("<p>Soll das Backup „%1“ wirklich wiederhergestellt werden?</p>"
               "<p>Die Turnierdatenbank „%2“ wird dabei überschrieben und etwaige Änderungen "
               "gehen verloren!</p>").arg(backupInfo.fileName(), dbToOverwriteInfo.fileName()),
            QMessageBox::Ok | QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    updateStatus(tr("Stelle Backup wieder her …"));

    if (! backupInfo.isReadable()) {
        QApplication::restoreOverrideCursor();
        updateStatus(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
               "<p>Die Datei „%1“ ist nicht lesbar. Bitte die Zugriffsrechte überprüfen!</p>").arg(
               QDir::toNativeSeparators(dialog.backupToRestore())));
        return;
    }

    const QString openedDb = m_db->getDbFile();
    closeTournament();

    // Is the backup we want to restore a valid database?

    if (m_db->checkDatabase(dialog.backupToRestore()) != Database::Error::NoError) {
        QApplication::restoreOverrideCursor();
        updateStatus(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>")
            + m_db->errorMessage()
            + tr("<p>Es wird jetzt versucht, die zuletzt geöffnete Datenbank wieder zu öffnen."
                 "</p>"));

        // Try to reload the database
        loadTournament(openedDb);
        return;
    }

    // To be sure, create a backup of the database to be overwritten

    // Get a safe destination file name
    int i = 0;
    QString tempDbBackup = dialog.dbToOverwrite() + QString::fromUtf8("~");
    while (QFile::exists(tempDbBackup)) {
        i++;
        tempDbBackup = dialog.dbToOverwrite() + QString::fromUtf8("~") + QString::number(i);
    }

    // Move the file to a backup name
    if (! QFile::rename(dialog.dbToOverwrite(), tempDbBackup)) {
        QApplication::restoreOverrideCursor();
        updateStatus(tr("Wiederherstellung fehlgeschlagen!"));
        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
               "<p>Die Datenbank „%1“ konnte nicht nach „%2“ umbenannt werden, daher wird der "
               "Vorgang sicherheitshalber abgebrochen.</p>"
               "<p>Es wird jetzt versucht, die zuletzt geöffnete Datenbank wieder zu öffnen."
               "</p>").arg(
               QDir::toNativeSeparators(dialog.dbToOverwrite()),
               QDir::toNativeSeparators(tempDbBackup)));

        // Try to reload the database
        loadTournament(openedDb);
        return;
    }

    // Copy the backup to restore
    if (! QFile::copy(dialog.backupToRestore(), dialog.dbToOverwrite())) {
        QApplication::restoreOverrideCursor();
        updateStatus(tr("Wiederherstellung fehlgeschlagen!"));

        QMessageBox::critical(this,
            tr("Backup wiederherstellen"),
            tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
               "<p>Das Backup „%1“ konnte nicht nach „%2“ kopiert werden!</p>"
               "<p>Es wird jetzt versucht, die Sicherungskopie der Datenbank wiederherzustellen "
               "und die Datenbank zu öffnen.</p>").arg(
               QDir::toNativeSeparators(dialog.backupToRestore()),
               QDir::toNativeSeparators(dialog.dbToOverwrite())));

        if (! QFile::rename(tempDbBackup, dialog.dbToOverwrite())) {
            // This one really should not happen ...
            QMessageBox::critical(this,
                tr("Backup wiederherstellen"),
                tr("<p><b>Wiederherstellung fehlgeschlagen</b></p>"
                   "<p>Die Sicherungskopie „%1“ der Datenbank konnte nicht nach „%2“ umbenannt "
                   "werden.</p>"
                   "<p>Die Sicherungskopie müsste vorhanden sein und sollte manuell umbenannt "
                   "werden können, um die Datenbank wiederherzustellen. Bitte überprüfen!"
                   "</p>").arg(
                QDir::toNativeSeparators(tempDbBackup),
                QDir::toNativeSeparators(dialog.dbToOverwrite())));
            return;
        }

        loadTournament(dialog.dbToOverwrite());
        return;
    }

    // Everything is fine :-)

    loadTournament(dialog.dbToOverwrite());
    QFileInfo tempBackupInfo(tempDbBackup);
    QApplication::restoreOverrideCursor();
    updateStatus(tr("Backup erfolgreich wiederhergestellt!"));

    QMessageBox::information(this,
        tr("Backup wiederherstellen"),
        tr("Das Backup „%1“ wurde erfolgreich wiederhergestellt!").arg(backupInfo.fileName()));

    if (! QFile::remove(tempDbBackup)) {
        // This should not happen
        QMessageBox::warning(this,
            tr("Backup wiederherstellen"),
            tr("Die temporäre Sicherungskopie „%1“, die vor dem Überschreiben der Datenbank "
               "angelegt wurde, konnte nicht gelöscht werden. Bitte manuell entfernen.").arg(
               QDir::toNativeSeparators(tempDbBackup)));
    }
}
