/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "PlayersPage.h"
#include "Database.h"
#include "Cell.h"
#include "SearchWidget.h"
#include "TableDelegate.h"
#include "DefinedCharacters.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QScrollBar>
#include <QMenu>
#include <QAction>
#include <QMessageBox>
#include <QApplication>
#include <QWidgetAction>
#include <QResizeEvent>

PlayersPage::PlayersPage(QWidget *parent, Database *db)
    : QWidget(parent), m_db(db)
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    // Register new players/pairs

    m_newPlayersBox = new QGroupBox;
    QHBoxLayout *newPlayersBoxLayout = new QHBoxLayout(m_newPlayersBox);
    layout->addWidget(m_newPlayersBox);

    m_nameLabel = new QLabel;
    m_nameLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    newPlayersBoxLayout->addWidget(m_nameLabel);

    m_playersName = new QLineEdit;
    newPlayersBoxLayout->addWidget(m_playersName);
    connect(m_playersName, &QLineEdit::textChanged, this, &PlayersPage::checkEnteredPlayersName);
    connect(m_playersName, &QLineEdit::returnPressed, this, &PlayersPage::registerPlayers);

    m_registerButton = new QPushButton(tr("Anmelden"));
    m_registerButton->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_registerButton->setEnabled(false);
    newPlayersBoxLayout->addWidget(m_registerButton);
    connect(m_registerButton, &QPushButton::clicked, this, &PlayersPage::registerPlayers);

    // Registered players/pairs

    m_playersBox = new QGroupBox;
    QVBoxLayout *playersBoxLayout = new QVBoxLayout(m_playersBox);
    layout->addWidget(m_playersBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    playersBoxLayout->addWidget(m_isSearchLabel);

    m_playersList = new QTableWidget(0, 1);
    TableDelegate *tableDelegate = new TableDelegate(this);
    m_playersList->setItemDelegate(tableDelegate);
    connect(tableDelegate, &TableDelegate::editEntered, this, &PlayersPage::editEntered);
    m_playersList->horizontalHeader()->hide();
    m_playersList->verticalHeader()->hide();
    m_playersList->setShowGrid(false);
    m_playersList->setWordWrap(false);
    m_playersList->setSelectionMode(QTableWidget::NoSelection);
    m_playersList->setFocusPolicy(Qt::NoFocus);
    m_playersList->setHorizontalScrollMode(QAbstractItemView::ScrollPerPixel);
    connect(m_playersList->verticalScrollBar(), &QScrollBar::valueChanged,
            m_playersList, &QTableWidget::resizeColumnsToContents);
    m_playersList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_playersList, &QTableWidget::customContextMenuRequested,
            this, &PlayersPage::showContextMenu);

    playersBoxLayout->addWidget(m_playersList);

    // Bottom line and Search widget

    QHBoxLayout *bottomLayout = new QHBoxLayout;
    playersBoxLayout->addLayout(bottomLayout);

    m_stateLabel = new QLabel;
    bottomLayout->addWidget(m_stateLabel);

    m_separatorLabel = new QLabel;
    bottomLayout->addWidget(m_separatorLabel);

    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &PlayersPage::updateSearch);
    bottomLayout->addWidget(m_searchWidget);

    // Context menu

    m_contextMenu = new QMenu(this);

    m_menuTitleAction = new QWidgetAction(this);
    m_menuTitle = new QLabel;
    m_menuTitle->setAlignment(Qt::AlignCenter);
    m_menuTitle->setMargin(4);
    m_menuTitleAction->setDefaultWidget(m_menuTitle);
    m_contextMenu->addAction(m_menuTitleAction);

    m_contextMenu->addSeparator();

    m_editAction = new QAction(this);
    connect(m_editAction, &QAction::triggered, this, &PlayersPage::editPlayers);
    m_contextMenu->addAction(m_editAction);

    m_deleteAction = new QAction(this);
    connect(m_deleteAction, &QAction::triggered, this, &PlayersPage::deletePlayers);
    m_contextMenu->addAction(m_deleteAction);

    m_contextMenu->addSeparator();

    m_markAction = new QAction(tr("Markieren"), this);
    connect(m_markAction, &QAction::triggered, this, &PlayersPage::markPlayers);
    m_contextMenu->addAction(m_markAction);

    m_topMarkAction = new QAction(tr("Markieren und oben einsortieren"), this);
    connect(m_topMarkAction, &QAction::triggered, this, &PlayersPage::topMarkPlayers);
    m_contextMenu->addAction(m_topMarkAction);

    m_unMarkAction = new QAction(tr("Markierung entfernen"), this);
    connect(m_unMarkAction, &QAction::triggered, this, &PlayersPage::unMarkPlayers);
    m_contextMenu->addAction(m_unMarkAction);

    m_contextMenu->addSeparator();

    QMenu *wholeListMenu = new QMenu(tr("Ganze Liste"), this);
    m_contextMenu->addMenu(wholeListMenu);

    m_markWholeListAction = new QAction(tr("Markieren"), this);
    connect(m_markWholeListAction, &QAction::triggered, this, &PlayersPage::markWholeList);
    wholeListMenu->addAction(m_markWholeListAction);

    m_topMarkWholeListAction = new QAction(tr("Markieren und oben einsortieren"), this);
    connect(m_topMarkWholeListAction, &QAction::triggered, this, &PlayersPage::topMarkWholeList);
    wholeListMenu->addAction(m_topMarkWholeListAction);

    m_unMarkWholeListAction = new QAction(tr("Markierung entfernen"), this);
    connect(m_unMarkWholeListAction, &QAction::triggered, this, &PlayersPage::unMarkWholeList);
    wholeListMenu->addAction(m_unMarkWholeListAction);

    wholeListMenu->addSeparator();

    m_deleteAllAction = new QAction(this);
    connect(m_deleteAllAction, &QAction::triggered, this, &PlayersPage::deleteAll);
    wholeListMenu->addAction(m_deleteAllAction);

    m_deleteMarkedAction = new QAction(this);
    connect(m_deleteMarkedAction, &QAction::triggered, this, &PlayersPage::deleteMarked);
    wholeListMenu->addAction(m_deleteMarkedAction);

    m_deleteUnMarkedAction = new QAction(this);
    connect(m_deleteUnMarkedAction, &QAction::triggered, this, &PlayersPage::deleteUnMarked);
    wholeListMenu->addAction(m_deleteUnMarkedAction);
}

void PlayersPage::setPlayersNameFocus()
{
    m_playersName->setFocus();
}

void PlayersPage::reload()
{
    m_newPlayersBox->show();
    m_newPlayersBox->setTitle(tr("Neue %1 anmelden").arg(
                                 m_db->playersString(Database::Numerus::plural)));
    m_playersBox->setTitle(tr("Angemeldete %1").arg(
                              m_db->playersString(Database::Numerus::plural)));
    m_deleteAction->setText(tr("%1 löschen").arg(m_db->playersString(Database::Numerus::singular)));
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        m_nameLabel->setText(tr("Name(n):"));
        m_editAction->setText(tr("Name(n) ändern"));
        m_deleteAllAction->setText(tr("Alle Paare löschen"));
        m_deleteMarkedAction->setText(tr("Alle markierten Paare löschen"));
        m_deleteUnMarkedAction->setText(tr("Alle nicht markierten Paare löschen"));
        break;
    case Database::TournamentMode::singlePlayers:
        m_nameLabel->setText(tr("Name:"));
        m_editAction->setText(tr("Name ändern"));
        m_deleteAllAction->setText(tr("Alle Spieler löschen"));
        m_deleteMarkedAction->setText(tr("Alle markierten Spieler löschen"));
        m_deleteUnMarkedAction->setText(tr("Alle nicht markierten Spieler löschen"));
        break;
    }

    m_playersName->setFocus();
    m_playersList->clear();

    int playersCount = 0;

    QStringList allPlayers = m_db->listPlayers();
    m_db->sortList(allPlayers);

    m_playersList->clear();
    m_playersList->setRowCount(allPlayers.size());
    int rowType = 1;

    m_playersList->blockSignals(true);

    QStringList topMarkedPlayers;
    QStringList playersWithoutTopMark;
    int topMarkedPlayersCount = 0;
    for (const QString &player : allPlayers) {
        if (player.startsWith(MARKER_CHARACTER)) {
            topMarkedPlayers << player;
            topMarkedPlayersCount++;
        } else {
            playersWithoutTopMark << player;
        }
    }
    allPlayers.clear();
    allPlayers << topMarkedPlayers << playersWithoutTopMark;

    int allMarkedPlayersCount = 0;
    for (const QString &player : allPlayers) {
        rowType++;
        rowType = rowType % 2;
        Qt::GlobalColor textColor = Qt::color0;
        if (player.contains(MARKER_CHARACTER)) {
            textColor = Qt::red;
            allMarkedPlayersCount++;
        }

        QTableWidgetItem *playerItem = new Cell(player, rowType, textColor);
        playerItem->setData(Qt::UserRole, QVariant(m_searchWidget->searchVariants(player)));
        m_playersList->setItem(playersCount, 0, playerItem);

        playersCount++;
    }

    m_unMarkWholeListAction->setEnabled(allMarkedPlayersCount > 0);
    m_topMarkWholeListAction->setEnabled(topMarkedPlayersCount < playersCount);
    m_markWholeListAction->setEnabled(allMarkedPlayersCount - topMarkedPlayersCount < playersCount);
    if (! m_db->tournamentStarted()) {
        m_deleteAllAction->setEnabled(playersCount > 0);
        m_deleteMarkedAction->setEnabled(allMarkedPlayersCount > 0);
        m_deleteUnMarkedAction->setEnabled(allMarkedPlayersCount < playersCount);
    } else {
        m_deleteAllAction->setEnabled(false);
        m_deleteMarkedAction->setEnabled(false);
        m_deleteUnMarkedAction->setEnabled(false);
    }

    m_playersList->resizeColumnsToContents();

    m_playersList->blockSignals(false);

    if (playersCount > 0) {
        if (! m_searchWidget->isVisible()) {
            m_separatorLabel->show();
            m_searchWidget->show();
        }
        emit enableExport(true);
    } else {
        m_separatorLabel->hide();
        m_searchWidget->hide();
        emit enableExport(false);
    }

    QString state;
    if (playersCount == 0) {
        state = tr("Keine %1 angemeldet").arg(m_db->playersString(Database::Numerus::plural));
    } else if (playersCount == 1) {
        state = tr("1 %1 angemeldet").arg(m_db->playersString(Database::Numerus::singular));
    } else {
        state = tr("%1 %2 angemeldet").arg(QString::number(playersCount),
                                           m_db->playersString(Database::Numerus::plural));
        if (allMarkedPlayersCount > 0) {
            if (allMarkedPlayersCount == playersCount) {
                state.append(tr(" (alle markiert)"));
            } else {
                state.append(tr(" (%1 markiert, %2 nicht)").arg(
                                QString::number(allMarkedPlayersCount),
                                QString::number(playersCount - allMarkedPlayersCount)));
            }
        }
    }

    if (playersCount % m_db->necessaryDivisor() != 0) {
        state.append(tr(" – <span style=\"color:red;\">nicht durch %1 teilbar!</span>").arg(
                        QString::number(m_db->necessaryDivisor())));
        emit tournamentCanBeStarted(false);
    } else {
        if (playersCount > 0) {
            emit tournamentCanBeStarted(true);
        }
    }

    m_stateLabel->setText(state);

    if (m_db->tournamentStarted()) {
        m_newPlayersBox->hide();
    }

    m_newPlayersBox->setEnabled(m_db->isWritable());
    if (m_db->isWritable()) {
        m_playersList->setEditTriggers(QAbstractItemView::DoubleClicked);
    } else {
        m_playersList->setEditTriggers(QAbstractItemView::NoEditTriggers);
    }

    scoreChanged();
}

void PlayersPage::reset()
{
    m_playersList->clear();
    m_newPlayersBox->show();
}

void PlayersPage::checkEnteredPlayersName(const QString &name)
{
    // Checks if a player/pair name is empty and enables/disables the 'register' button
    m_registerButton->setEnabled(! name.simplified().isEmpty());
}

void PlayersPage::registerPlayers()
{
    if (! m_registerButton->isEnabled()) {
        return;
    }

    QString name = m_playersName->text().simplified();

    if (! m_db->checkName(name)) {
        QMessageBox::critical(this,
                              m_newPlayersBox->title(),
                              tr("Der Name „%1“ ist bereits vergeben!").arg(name),
                              QMessageBox::Ok);
        return;
    }

    emit statusUpdate(tr("Melde %1 an …").arg(m_db->playersString(Database::Numerus::singular)));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_db->registerPlayers(name);
    m_playersName->clear();
    m_playersName->setFocus();
    reload();
    m_searchWidget->clear();

    // Make sure the entered name is visible
    int targetRow = 0;
    for (int i = 0; i < m_playersList->rowCount(); i++) {
        if (m_playersList->item(i, 0)->text() == name) {
            targetRow = i;
            break;
        }
    }
    if (targetRow == m_playersList->rowCount() - 1) {
        m_playersList->scrollToBottom();
    } else {
        m_playersList->scrollToItem(m_playersList->item(targetRow, 0),
                                    QAbstractItemView::PositionAtCenter);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("%1 „%2“ angemeldet").arg(
                         m_db->playersString(Database::Numerus::singular), name));
}

void PlayersPage::editPlayers()
{
    m_playersList->editItem(m_playersList->currentItem());
}

void PlayersPage::editEntered(const QString &originalName, const QString &name)
{
    if (! m_db->checkName(name)) {
        QMessageBox::critical(this,
            tr("%1 umbenennen nicht möglich").arg(m_db->playersString(Database::Numerus::singular)),
            tr("Der Name „%1“ ist bereits vergeben!").arg(name),
            QMessageBox::Ok);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Benenne %1 um …").arg(m_db->playersString(Database::Numerus::singular)));
    m_db->renamePlayers(originalName, name);
    emit statusUpdate(tr("%1 „%2“ umbenannt zu „%3“").arg(
                         m_db->playersString(Database::Numerus::singular), originalName, name));
    QApplication::restoreOverrideCursor();
    listModified();
}

void PlayersPage::scoreChanged()
{
    m_deleteAction->setEnabled(! m_db->tournamentStarted());
}

void PlayersPage::showContextMenu(const QPoint &point)
{
    if (! m_db->isWritable()) {
        return;
    }

    QTableWidgetItem *selectedPlayers = m_playersList->currentItem();
    if (selectedPlayers == nullptr) {
        return;
    }

    m_menuTitle->setText(tr("%1 „%2“").arg(m_db->playersString(Database::Numerus::singular),
                                           selectedPlayers->text()));
    QResizeEvent resizeEvent(m_menuTitle->size(), m_contextMenu->size());
    QApplication::sendEvent(m_contextMenu, &resizeEvent);

    if (! selectedPlayers->text().contains(MARKER_CHARACTER)) {
        m_unMarkAction->setEnabled(false);
        m_markAction->setEnabled(true);
        m_topMarkAction->setEnabled(true);
    } else if (selectedPlayers->text().startsWith(MARKER_CHARACTER)) {
        m_unMarkAction->setEnabled(true);
        m_markAction->setEnabled(true);
        m_topMarkAction->setEnabled(false);
    } else if (selectedPlayers->text().endsWith(MARKER_CHARACTER)) {
        m_unMarkAction->setEnabled(true);
        m_markAction->setEnabled(false);
        m_topMarkAction->setEnabled(true);
    }

    m_contextMenu->exec(m_playersList->mapToGlobal(point));
}

void PlayersPage::deletePlayers()
{
    QString toDelete = m_playersList->currentItem()->text();

    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("%1 löschen").arg(
                                 m_db->playersString(Database::Numerus::singular)));
    messageBox.setText(tr("Soll %1 „%2“ wirklich gelöscht werden?").arg(
                          m_db->playersString(Database::Numerus::singular, true), toDelete));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    m_db->deletePlayers(toDelete);
    listModified();
}

void PlayersPage::updateSearch(QString searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_playersList->rowCount(); row++) {
            if (m_playersList->isRowHidden(row)) {
                m_playersList->showRow(row);
            }
        }
        m_playersList->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    int matches = 0;
    const QStringList searchVariants = m_searchWidget->searchVariants(searchText);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (m_searchWidget->checkMatch(
                m_playersList->item(row, 0)->data(Qt::UserRole).toStringList(),
                searchVariants)
           ) {
            if (m_playersList->isRowHidden(row)) {
                m_playersList->showRow(row);
            }
            matches++;
        } else {
            if (! m_playersList->isRowHidden(row)) {
                m_playersList->hideRow(row);
            }
        }
    }
    m_searchWidget->updateMatches(matches);
    m_playersList->resizeColumnsToContents();

    if (! m_isSearchLabel->isVisible()) {
        m_isSearchLabel->show();
    }
    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText));
}

void PlayersPage::markPlayers()
{
    m_db->renamePlayers(m_playersList->currentItem()->text(),
                        m_playersList->currentItem()->text().remove(MARKER_CHARACTER)
                        + MARKER_CHARACTER);
    listModified();
}

void PlayersPage::topMarkPlayers()
{
    m_db->renamePlayers(m_playersList->currentItem()->text(),
                        MARKER_CHARACTER
                        + m_playersList->currentItem()->text().remove(MARKER_CHARACTER));
    listModified();
}

void PlayersPage::unMarkPlayers()
{
    m_db->renamePlayers(m_playersList->currentItem()->text(),
                        m_playersList->currentItem()->text().remove(MARKER_CHARACTER));
    listModified();
}

void PlayersPage::markWholeList()
{
    QVariantList oldPlayerNames;
    QVariantList newPlayerNames;
    for (const QString &oldName : m_db->listPlayers()) {
        QString newName = oldName;
        newName.remove(MARKER_CHARACTER);
        newName += MARKER_CHARACTER;
        if (newName != oldName) {
            oldPlayerNames << oldName;
            newPlayerNames << newName;
        }
    }
    m_db->renamePlayers(oldPlayerNames, newPlayerNames);
    listModified();
}

void PlayersPage::topMarkWholeList()
{
    QVariantList oldPlayerNames;
    QVariantList newPlayerNames;
    for (const QString &oldName : m_db->listPlayers()) {
        QString newName = oldName;
        newName.remove(MARKER_CHARACTER);
        newName = MARKER_CHARACTER + newName;
        if (newName != oldName) {
            oldPlayerNames << oldName;
            newPlayerNames << newName;
        }
    }
    m_db->renamePlayers(oldPlayerNames, newPlayerNames);
    listModified();
}

void PlayersPage::unMarkWholeList()
{
    QVariantList oldPlayerNames;
    QVariantList newPlayerNames;
    for (const QString &oldName : m_db->listPlayers()) {
        QString newName = oldName;
        newName.remove(MARKER_CHARACTER);
        if (newName != oldName) {
            oldPlayerNames << oldName;
            newPlayerNames << newName;
        }
    }
    m_db->renamePlayers(oldPlayerNames, newPlayerNames);
    listModified();
}

void PlayersPage::deleteMarked()
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("%1 löschen").arg(
                                 m_db->playersString(Database::Numerus::plural)));
    messageBox.setText(tr("Sollen wirklich alle markierten %1 gelöscht werden?").arg(
                          m_db->playersString(Database::Numerus::plural)));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    QVariantList toDelete;
    for (const QString &name : m_db->listPlayers()) {
        if (name.contains(MARKER_CHARACTER)) {
            toDelete << name;
        }
    }

    m_db->deletePlayers(toDelete);
    listModified();
}

void PlayersPage::deleteUnMarked()
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("%1 löschen").arg(
                                 m_db->playersString(Database::Numerus::plural)));
    messageBox.setText(tr("Sollen wirklich alle nicht markierten %1 gelöscht werden?").arg(
                          m_db->playersString(Database::Numerus::plural)));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    QVariantList toDelete;
    for (const QString &name : m_db->listPlayers()) {
        if (! name.contains(MARKER_CHARACTER)) {
            toDelete << name;
        }
    }

    m_db->deletePlayers(toDelete);
    listModified();
}

void PlayersPage::deleteAll()
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("%1 löschen").arg(
                                 m_db->playersString(Database::Numerus::plural)));
    messageBox.setText(tr("Sollen wirklich alle %1 gelöscht werden?").arg(
                          m_db->playersString(Database::Numerus::plural)));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    m_db->reset();
    reload();
    m_searchWidget->clear();
}

void PlayersPage::listModified()
{
    emit playersListChanged();
    m_searchWidget->saveSearch();
    reload();
    m_searchWidget->restoreSearch();
}
