/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "SearchWidget.h"
#include "SearchLineEdit.h"

// Qt includes
#include <QHBoxLayout>
#include <QLabel>

namespace {

const QList<QChar> CHARS_TO_REPLACE({
    QChar(0x00e4), // ä
    QChar(0x00f6), // ö
    QChar(0x00fc), // ü
    QChar(0x00df), // ß
    QChar(0x00E6), // æ
    QChar(0x0153), // œ
    QChar(0x00F8)  // ø
});

const QStringList CHAR_REPLACEMENTS({
    QString::fromUtf8("ae"),
    QString::fromUtf8("oe"),
    QString::fromUtf8("ue"),
    QString::fromUtf8("ss"),
    QString::fromUtf8("ae"),
    QString::fromUtf8("oe"),
    QString::fromUtf8("oe")
});

}

SearchWidget::SearchWidget(QWidget *parent) : QWidget(parent)
{
    QHBoxLayout *layout = new QHBoxLayout(this);

    m_searchLabel = new QLabel(tr("Suche:"));
    layout->addWidget(m_searchLabel);

    m_search = new SearchLineEdit;
    m_search->setClearButtonEnabled(true);
    connect(m_search, &QLineEdit::textChanged, this, &SearchWidget::processInput);
    layout->addWidget(m_search);

    m_matches = new QLabel;
    m_matches->hide();
    layout->addWidget(m_matches);

    m_locale = QLocale();
}

void SearchWidget::updateMatches(int matches)
{
    if (! m_matches->isVisible()) {
        m_matches->show();
    }

    if (matches == 0) {
        m_matches->setText(tr("<span style=\"color:red\">Keine Treffer</span>"));
    } else {
        m_matches->setText(tr("%1 Treffer").arg(QString::number(matches)));
    }
}

void SearchWidget::processInput(const QString &text)
{
    emit searchChanged(text);
    if (text.isEmpty()) {
        m_matches->hide();
    }
}

void SearchWidget::clear()
{
    m_search->clear();
    m_savedSearch.clear();
}

void SearchWidget::saveSearch()
{
    m_savedSearch = m_search->text();
    m_search->clear();
}

void SearchWidget::restoreSearch()
{
    if (! m_savedSearch.isEmpty()) {
        m_search->setText(m_savedSearch);
    }
}

QStringList SearchWidget::searchVariants(const QString &searchString) const
{
    QStringList variants;

    // The "normal" search variant is just a lowercase version of the actual string
    QString lower = m_locale.toLower(searchString);
    variants << lower;

    // We replace some special characters (e. g. German umlauts) here
    QString replaced = lower;
    for (int i = 0; i < CHARS_TO_REPLACE.size(); i++) {
        replaced.replace(CHARS_TO_REPLACE.at(i), CHAR_REPLACEMENTS.at(i));
    }
    if (! variants.contains(replaced)) {
        variants << replaced;
    }

    // Here, we strip out all diacritical marks and accents from both versions by
    // first separating the combinig characters and then removing them
    // cf. http://en.wikipedia.org/wiki/Combining_character
    lower = lower.normalized(QString::NormalizationForm_D);
    replaced = replaced.normalized(QString::NormalizationForm_D);
    for(int i = 0x0300; i <= 0x036F; i++) {
        lower.remove(QChar(i));
        replaced.remove(QChar(i));
    }
    if (! variants.contains(lower)) {
        variants << lower;
    }
    if (! variants.contains(replaced)) {
        variants << replaced;
    }

    return variants;
}

bool SearchWidget::checkMatch(const QStringList &textVariants,
                              const QStringList &searchVariants) const
{
    for (const QString &searchVariant : searchVariants) {
        for (const QString &textVariant : textVariants) {
            if (textVariant.contains(searchVariant)) {
                return true;
            }
        }
    }
    return false;
}
