/* Copyright (c) 2017-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "StopWatch.h"
#include "TimeLabel.h"

// Qt includes
#include <QDebug>
#include <QTimer>
#include <QVBoxLayout>
#include <QLabel>
#include <QHBoxLayout>
#include <QSpinBox>
#include <QPushButton>
#include <QCheckBox>
#include <QLineEdit>

StopWatch::StopWatch(QWidget *parent) : QDialog(parent)
{
    setAttribute(Qt::WA_DeleteOnClose);
    setWindowTitle(tr("Muckturnier: Stoppuhr"));

    m_timer = new QTimer(this);
    m_timer->setTimerType(Qt::PreciseTimer);
    connect(m_timer, &QTimer::timeout, this, &StopWatch::update);

    QVBoxLayout *layout = new QVBoxLayout(this);

    m_title = new ExpandingLabel;
    m_title->hide();
    layout->addWidget(m_title);

    m_timeDisplay = new TimeLabel;
    m_timeDisplay->hide();
    layout->addWidget(m_timeDisplay);

    m_settingsWidget = new QWidget;
    QVBoxLayout *settingsLayout = new QVBoxLayout(m_settingsWidget);

    m_displayTitle = new QCheckBox(tr("Titel anzeigen"));
    settingsLayout->addWidget(m_displayTitle);
    connect(m_displayTitle, &QCheckBox::stateChanged, this, &StopWatch::toggleTitleInput);
    m_titleInput = new QLineEdit;
    m_titleInput->hide();
    settingsLayout->addWidget(m_titleInput);

    QHBoxLayout *timeSpanLayout = new QHBoxLayout;
    settingsLayout->addLayout(timeSpanLayout);

    timeSpanLayout->addStretch();
    m_hours = new QSpinBox;
    timeSpanLayout->addWidget(m_hours);
    timeSpanLayout->addWidget(new QLabel(tr("h")));
    m_minutes = new QSpinBox;
    timeSpanLayout->addWidget(m_minutes);
    timeSpanLayout->addWidget(new QLabel(tr("m")));
    m_seconds = new QSpinBox;
    timeSpanLayout->addWidget(m_seconds);
    timeSpanLayout->addWidget(new QLabel(tr("s")));
    timeSpanLayout->addStretch();

    QPushButton *start = new QPushButton(tr("Start"));
    connect(start, &QPushButton::clicked, this, &StopWatch::start);
    settingsLayout->addWidget(start);

    layout->addWidget(m_settingsWidget);

    m_controlWidget = new QWidget;
    m_controlWidget->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    QHBoxLayout *controlLayout = new QHBoxLayout(m_controlWidget);

    m_pause = new QPushButton(tr("Pause"));
    m_pause->setCheckable(true);
    connect(m_pause, &QPushButton::clicked, this, &StopWatch::pause);
    controlLayout->addWidget(m_pause);

    QPushButton *stop = new QPushButton(tr("Stop"));
    connect(stop, &QPushButton::clicked, this, &StopWatch::stop);
    controlLayout->addWidget(stop);

    m_controlWidget->hide();
    layout->addWidget(m_controlWidget);

    m_isRunning = false;
}

void StopWatch::setTarget(qint64 seconds)
{
    m_targetTime = QDateTime::currentDateTime().addSecs(seconds);
}

void StopWatch::start()
{
    setTarget(m_hours->value() * 3600 + m_minutes->value() * 60 + m_seconds->value());
    m_settingsWidget->hide();
    m_timeDisplay->show();

    if (m_displayTitle->isChecked()) {
        QString title = m_titleInput->text().simplified();
        if (! title.isEmpty()) {
            m_title->setText(title);
            m_title->show();
            m_title->resizeText();
        }
    }

    m_controlWidget->show();
    m_timer->start(1000);
    update();
    m_timeDisplay->resizeText();

    m_isRunning = true;
}

void StopWatch::update()
{
    QDateTime currentDateTime = QDateTime::currentDateTime();
    // We use msecsTo here instead of secsTo to avoid rounding problems.
    // If we use secsTo, we get "0" both for the currentDateTime == m_targetTime
    // and also for one second before.
    m_timeDisplay->setSeconds(qRound64(currentDateTime.msecsTo(m_targetTime) / 1000.0));
}

void StopWatch::stop()
{
    m_timer->stop();
    m_title->hide();
    m_timeDisplay->hide();
    m_controlWidget->hide();
    m_settingsWidget->show();
    if (m_pause->isChecked()) {
        m_pause->setChecked(false);
    }

    m_isRunning = false;
}

void StopWatch::pause()
{
    if (m_pause->isChecked()) {
        m_timer->stop();
    } else {
        setTarget(m_timeDisplay->seconds());
        m_timer->start(1000);
        m_pause->clearFocus();
    }
}

void StopWatch::toggleTitleInput(bool state)
{
    if (state) {
        m_titleInput->show();
        m_titleInput->setFocus();
    } else {
        m_titleInput->hide();
    }
}

bool StopWatch::isRunning() const
{
    return m_isRunning;
}
