/* Copyright (c) 2012-2018 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "TournamentInfoDialog.h"
#include "Database.h"
#include "DefinedCharacters.h"

// Qt includes
#include <QGridLayout>
#include <QLabel>
#include <QDir>
#include <QGroupBox>
#include <QDebug>
#include <QCryptographicHash>

TournamentInfoDialog::TournamentInfoDialog(QWidget *parent, Database *database)
    : Dialog(parent, database)
{
    setTitle(tr("Turnierinformationen"));
    showOnlyOkButton();

    addGroup(tr("Datenbank"));

    addLabel(tr("Dateiname:"));
    QString dbFile = QDir::toNativeSeparators(db()->getDbFile());
    QFont font = this->font();
    QFontMetrics fontMetrics(font);
    if (fontMetrics.boundingRect(dbFile).width() > fontMetrics.boundingRect(
        QString::fromUtf8("/A/very/long/path/which/still/does/not/mess/up/the/layout")).width()) {
        addInfo(dbFile, true);
    } else {
        addInfo(dbFile, false);
    }

    addLabel(tr("Erstellt mit:"));
    addInfo(tr("Muckturnier %1").arg(db()->getMuckturnierVersion()));

    addGroup(tr("Einstellungen"));

    addLabel(tr("Tuniermodus:"));
    switch (db()->getTournamentMode()) {
    case Database::fixedPairs:
        addInfo(tr("Feste Paare"));
        break;
    case Database::singlePlayers:
        addInfo(tr("Einzelne Spieler"));
        break;
    }

    addLabel(tr("Punkte pro Bobbl:"));
    addInfo(QString::number(db()->getBoogerScore()));

    addLabel(tr("Bobbl pro Runde:"));
    addInfo(QString::number(db()->getBoogersPerRound()));

    addGroup(tr("Turnier"));

    addLabel(tr("Angemeldete %1:").arg(db()->playersString(Database::Numerus::plural)));
    int playersCount = db()->playersCount();
    addInfo(QString::number(playersCount));

    addLabel(tr("Gespielte Runden:"));
    QMap<int, bool> roundsState = db()->roundsState();
    int rounds = roundsState.count();
    if (rounds == 1) {
        QList<Database::Score> scores = db()->getScores(1);
        if (scores.count() == 0) {
            rounds = 0;
        }
    }
    addInfo(QString::number(rounds));

    bool allFinished = true;
    for (const bool &finished : roundsState.values()) {
        if (! finished) {
            allFinished = false;
            break;
        }
    }
    addLabel(tr("Alle Runden abgeschlossen:"));
    if (allFinished) {
        addInfo(tr("Ja"));
    } else {
        addInfo(tr("Nein"));
    }

    if (playersCount > 0) {

        addGroup(tr("Prüfsummen"));

        addLabel(tr("%1liste:").arg(db()->playersString(Database::Numerus::singular)));
        QStringList playersList = db()->listPlayers();
        for (int i = 0; i < playersList.size(); i++) {
            if (playersList[i].contains(MARKER_CHARACTER)) {
                playersList[i].remove(MARKER_CHARACTER);
            }
        }
        playersList.sort();
        addInfo(checksum(playersList.join(QString())));

        if (rounds > 0) {
            addLabel(tr("Rangliste:"));
            QList<Database::Ranking> allRankings = db()->getRanking(rounds);
            QString rankingData;
            for (Database::Ranking &ranking : allRankings) {
                if (ranking.name.contains(MARKER_CHARACTER)) {
                    ranking.name.remove(MARKER_CHARACTER);
                }
                rankingData += ranking.name
                            + QString::number(ranking.boogers)
                            + QString::number(ranking.goals)
                            + QString::number(ranking.opponentGoals);
            }
            addInfo(checksum(rankingData));
        }

    }
}

QString TournamentInfoDialog::checksum(const QString &text) const
{
    QString checksum = QString::fromUtf8(QCryptographicHash::hash(text.toUtf8(),
                                         QCryptographicHash::Md5).toHex());
    for (int i = 4; i < checksum.size(); i += 5) {
        checksum.insert(i, QString::fromUtf8(" "));
    }
    return QString::fromUtf8("<pre>%1</pre>").arg(checksum);
}

void TournamentInfoDialog::addGroup(const QString &text)
{
    QGroupBox *group = new QGroupBox(text);
    layout()->addWidget(group);
    QGridLayout *layout = new QGridLayout(group);
    m_currentGroupLayout = layout;
    m_row = -1;
}

void TournamentInfoDialog::addLabel(const QString &text)
{
    m_row++;
    QLabel *label = new QLabel(text);
    label->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Minimum);
    label->setTextInteractionFlags(Qt::TextBrowserInteraction | Qt::TextSelectableByMouse);
    label->setAlignment(Qt::AlignTop);
    m_currentGroupLayout->addWidget(label, m_row, 0);
}

void TournamentInfoDialog::addInfo(const QString &text, bool wordWrap)
{
    QLabel *label = new QLabel(text);
    label->setTextInteractionFlags(Qt::TextBrowserInteraction | Qt::TextSelectableByMouse);
    if (wordWrap) {
        label->setWordWrap(true);
    }
    m_currentGroupLayout->addWidget(label, m_row, 1);
}
